#include "main.h"



void board_init ()
{
    board * B;

    for (int i = 0; i < MAX_DEPTH; i ++)
    {
        B = &_B[i];

        B->check_is = 0;
        B->attacker = 0;
        B->castling_rights = 15;
        B->en_passant = 0;
        B->hash = E;
        B->hash_mat = E;
        B->taken = no_p;
        B->mat_key = 0;
        B->killer1 = B->killer2 = 0;
    };

    for (int j = 0; j < 12; j ++) P->piece [j] = E;
    P->occupied [0] = P->occupied [1] = E;
    for (int j = 0; j < 64; j ++) P->squares [j] = no_p;
};

void fen_to_pos (const char * fen, board * B)
{
	int x = A8, t = 0;
	U64 x64;
	int a;
	char ch;

	reps_refresh ();

	for (a = 0; a < 12; a ++) P->piece [a] = E;

	B->castling_rights = 0;
	B->en_passant = 0; // ,   
	B->attacker = 0;
	B->check_is = 0;
	B->taken = no_p;
	S->count_moves = 0;
	S->fifty = 0;
	B->hash = E;
	B->size = 0;
	S->irreversible_ply = 0;

	P->wtm = 1;

	B->mat_key = 0;


	////  
	for (;; fen ++)
	{
		if (x < 0) break;
		if (!(ch = *fen)) goto exit_f;
		if (ch == ' ') break;

		switch (ch)
		{
			case 'K': P->piece [WK] |= I << x; P->squares [x] = WK;                             x ++; t ++; break;
			case 'Q': P->piece [WQ] |= I << x; P->squares [x] = WQ; B->mat_key += mat_keys[WQ]; x ++; t ++; break;

			case 'R': P->piece [WR] |= I << x; P->squares [x] = WR; B->mat_key += mat_keys[WR]; x ++; t ++; break;
			case 'B': P->piece [WB] |= I << x; P->squares [x] = WB; B->mat_key += mat_keys[WB]; x ++; t ++; break;
			case 'N': P->piece [WN] |= I << x; P->squares [x] = WN; B->mat_key += mat_keys[WN]; x ++; t ++; break;
			case 'P': P->piece [WP] |= I << x; P->squares [x] = WP; B->mat_key += mat_keys[WP]; x ++; t ++; break;

			case 'k': P->piece [BK] |= I << x; P->squares [x] = BK;                             x ++; t ++; break;
			case 'q': P->piece [BQ] |= I << x; P->squares [x] = BQ; B->mat_key += mat_keys[BQ]; x ++; t ++; break;

			case 'r': P->piece [BR] |= I << x; P->squares [x] = BR; B->mat_key += mat_keys[BR]; x ++; t ++; break;
			case 'b': P->piece [BB] |= I << x; P->squares [x] = BB; B->mat_key += mat_keys[BB]; x ++; t ++; break;
			case 'n': P->piece [BN] |= I << x; P->squares [x] = BN; B->mat_key += mat_keys[BN]; x ++; t ++; break;
			case 'p': P->piece [BP] |= I << x; P->squares [x] = BP; B->mat_key += mat_keys[BP]; x ++; t ++; break;

			case '/': x ++; t ++; break;
			default : if (ch >= '1' && ch <= '8') {t += ch - '0'; x += ch - '0';}; break;
		};

		if (t > 8)
		{
			t = 0;
			x -= 0x10;
			x &= 0xF8;
		};
	};

	//     (   )
	if (pop_count (P->piece[WQ]) > 2) B->mat_key |= UNUSIAL_MAT;
	if (pop_count (P->piece[WR]) > 2) B->mat_key |= UNUSIAL_MAT;
	if (pop_count (P->piece[WB]) > 2) B->mat_key |= UNUSIAL_MAT;
	if (pop_count (P->piece[WN]) > 2) B->mat_key |= UNUSIAL_MAT;
	if (pop_count (P->piece[BQ]) > 2) B->mat_key |= UNUSIAL_MAT;
	if (pop_count (P->piece[BR]) > 2) B->mat_key |= UNUSIAL_MAT;
	if (pop_count (P->piece[BB]) > 2) B->mat_key |= UNUSIAL_MAT;
	if (pop_count (P->piece[BN]) > 2) B->mat_key |= UNUSIAL_MAT;

	//  occupied   
	P->occupied[0] = E;
	P->occupied[1] = E;

	for (a = 0; a < 12; a ++) P->occupied[a & 1] |= P->piece [a];

	if (!*fen) goto exit_f;

	////  

	//  
	fen ++;
	ch = *fen;
	if (ch == 'w') P->wtm = 1;
	else if (ch == 'b') P->wtm = 0;
	else if (!ch) goto exit_f;


	//   
	fen ++;
	if (!*fen) goto exit_f;
	fen ++;

	for (;; fen ++)
	{
		ch = *fen;
		if (ch == ' ' || ch == '-') break;
		if (!ch) goto exit_f;

		switch (ch)
		{
			case 'K': B->castling_rights |= c_K; break;
			case 'Q': B->castling_rights |= c_Q; break;
			case 'k': B->castling_rights |= c_k; break;
			case 'q': B->castling_rights |= c_q; break;
		};
	};


	//     

	for (;ch <= 'a' || ch >= 'h'; fen ++)
	{
		ch = *fen;
		if (!ch) goto exit_f;
	};

	a = ch - 'a';
	B->en_passant = a;

	ch = *fen;
	if (!ch) goto exit_f;
	if (ch >= '1' && ch <= '8') // ,      3,  6
	{
		a = ch - '1';
		B->en_passant += a << 3;
	};

	//    50- 
	if (!*(++fen)) goto exit_f;
	if (!*(++fen)) goto exit_f;

	for (; (ch = *fen); fen ++)
	{
		if (!ch) goto exit_f;
		if (ch >= '0' && ch <= '9')
		{
			S->fifty *= 10;
			S->fifty += ch - '0';
		}
		else break;
	};

	if (!*(fen++)) goto exit_f;

	//  
	for (; (ch = *fen); fen ++)
	{
		if (!ch) goto exit_f;
		if (ch >= '0' && ch <= '9')
		{
			S->count_moves *= 10;
			S->count_moves += ch - '0';
			S->irreversible_ply = S->count_moves;
		}
		else break;
	};

exit_f:

	R[0].key = B->hash = hash_calc (B);
};

char* pos_to_fen (board * B)
{
    int  i;
    int  x = A8, t = 0;
    int  p, c = 0;      //   
    //char str [25] = {0, };
    //char ch;
    char *pfen;
    U64  occ = P->occupied [0] | P->occupied [1];

    char *fen = new char [512];

    for (i = 0; i < 512; i ++) fen [i] = 0;

    pfen = fen;
    //pstr = str;

    for (;;)
    {
        if (occ & (I << x))
        {
            i = P->squares [x];
            if (c) *(pfen ++) = c + '0';
            *(pfen ++) = piece_array [i];
            c = 0;
        }
        else c ++;

        x ++;
        t ++;

        if (t >= 8)
        {
            t = 0;
			x -= 0x10;

			if (c)
			{
			    *(pfen ++) = c + '0';
                c = 0;
			};

			if (x < 0) break;

			*(pfen ++) = '/';
        };
    };

    //"rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"

    *(pfen ++) = ' ';
    *(pfen ++) = P->wtm ? 'w' : 'b';
    *(pfen ++) = ' ';

    p = B->castling_rights;

    if (p & c_K) *(pfen ++) = 'K';
    if (p & c_Q) *(pfen ++) = 'Q';
    if (p & c_k) *(pfen ++) = 'k';
    if (p & c_q) *(pfen ++) = 'q';

    if (p) *(pfen ++) = ' ';

    p = B->en_passant;

    if (!p) *(pfen ++) = '-';
    else
    {
        *(pfen ++) = vert (p) + 'a';
        *(pfen ++) = hori (p) + '1';
    };

    *(pfen ++) = ' ';

    p = S->count_moves;
    if (p < 0 || p > 500) p = 0;

    sprintf (pfen, "%i %i ", S->fifty, p);

    return fen;
};

U64 attacked_by (U64 & occ, int sq) //      ?
{
    U64 bb;
    U64 rq, bq;  //    
    U64 kn, ki; // , 

    bb = E;

    rq = P->piece[BR] | P->piece[WR] | P->piece[BQ] | P->piece [WQ];
    bq = P->piece[BB] | P->piece[WB] | P->piece[BQ] | P->piece [WQ];
    kn = P->piece[BN] | P->piece[WN];
    ki = P->piece[BK] | P->piece[WK];

    bb |= r_att (occ, sq) & rq;
    bb |= b_att (occ, sq) & bq;
    bb |= knight_att [sq] & kn;
    bb |= wpawn_att [sq]  & P->piece [BP];
    bb |= bpawn_att [sq]  & P->piece [WP];
    bb |= king_att [sq]   & ki;

    return bb;
};

// col -   
U64 attacked_by_col (U64 & occ, int sq, int col)
{
    U64 bb;
    U64 rq, bq;  //    

    rq = bq = P->piece[BQ + col];
    rq |= P->piece[BR + col];
    bq |= P->piece[BB + col];

    bb  = r_att (occ, sq) & rq;
    bb |= b_att (occ, sq) & bq;
    bb |= knight_att [sq] & P->piece[BN + col];
    bb |= king_att [sq]   & P->piece[BK + col];

    if (col) bb |= bpawn_att [sq] & P->piece [WP];
    else     bb |= wpawn_att [sq] & P->piece [BP];

    return bb;
};

/*
// col -   
U64 attack_table (U64 & occ, int sq, int col) //   ,   
{
    int i, j;
    U64 bb, p, att, occ_opp;
    int coeff[] = {-8, 8, -6, 6, -5, 5, -4, 4, -3, 3, -2, 2};

    occ_opp = P->occupied[col ^ 1];
    J->att_once = E;

    // 
    i = BK + col;
    bb = P->piece[i];
    j = bitscan (bb);
    att = king_att [j];

    // 
    i -= 2;
    for (bb = P->piece [i]; bb; bb ^= p)
    {
        p = LSB (bb);
        j = bitscan (p);

        att = q_att (occ, j);
    };

    // 
    i -= 2;
    for (bb = P->piece [i]; bb; bb ^= p)
    {
        p = LSB (bb);
        j = bitscan (p);

        att = r_att (occ, j);
    };

    // 
    i -= 2;
    for (bb = P->piece [i]; bb; bb ^= p)
    {
        p = LSB (bb);
        j = bitscan (p);

        att = b_att (occ, j);
    };

    // 
    i -= 2;
    for (bb = P->piece [i]; bb; bb ^= p)
    {
        p = LSB (bb);
        j = bitscan (p);

        att = knight_att [j];
    };

    // 
    i -= 2;
    bb = P->piece [i];

    if (!P->wtm) att = ( ((bb & ~file_h) >> 7) | ((bb & ~file_a) >> 9) ) & occ_opp;
    else         att = ( ((bb & ~file_h) << 9) | ((bb & ~file_a) << 7) ) & occ_opp;
};
*/

// col -   
int is_attacked (U64 & occ, int sq, int col)
{
    U64 rq, bq; //    

    if (knight_att [sq] & P->piece[BN + col]) return 1;

    rq = bq = P->piece[BQ + col];
    bq |= P->piece[BB + col];

    if (b_att (occ, sq) & bq) return 1;

    rq |= P->piece[BR + col];

    if (r_att (occ, sq) & rq) return 1;
    if (king_att [sq] & P->piece[BK + col]) return 1;

    if (col)
    {
        if (bpawn_att [sq] & P->piece [WP]) return 1;
    }
    else
    {
        if (wpawn_att [sq] & P->piece [BP]) return 1;
    };

    return 0;
};

int print_board (board * B)
{
    static const char castl [16][5] = {"-", "K", "Q", "QK", "k", "kK", "kQ", "kQK",
                                       "q", "qK", "qQ", "qQK", "qk", "qkK", "qkQ", "qkQK"};

	int a, b, c;
	U64 occ;

	occ = P->occupied[0] | P->occupied[1];

	for (a = 0; a <= 7; a ++)
	{
		uci_send ("\n");
		for (b = 0; b <= 7; b ++)
		{
			c = ((7 - a) << 3) + b;

            if (occ & (I << c)) uci_send ("%c", piece_array [P->squares[c]]);
            else                uci_send (".");
		};
	};
	uci_send ("\n");
	if (B->check_is) uci_send ("\ncheck!\n");

    uci_send ("\ncastling: %s\n", &castl [B->castling_rights][0]);

	return 0;
};

void build_mask (board * B)
{
    int coords_att, coords_king;
    U64 king;

    king = P->piece [BK + P->wtm];

    coords_att  = B->attacker;    //  
    coords_king = bitscan (king); // 

    //    
    B->msk_bb = link_att [coords_att][coords_king];
    B->msk_bb |= I << coords_att;
};

void check_detection (board * B, int fast)
{
    int j, k;
    U64 att, king;
    U64 occ, t;
    U64 rq, bq; //    
    U64 x64;   //  popcount'

    B->check_is = 0;
    B->attacker = 0;

    rq = bq = P->piece [BQ + !P->wtm];
    rq |= P->piece [BR + !P->wtm];
    bq |= P->piece [BB + !P->wtm];

    occ = P->occupied [0] | P->occupied [1];

    king = P->piece [BK + P->wtm];
    k = bitscan (king);

    //      
    att = (r_att (occ, k) & rq) | (b_att (occ, k) & bq);

    B->attacker = bitscan (att);
    B->check_is = pop_count (att);

    if (fast && B->check_is > 1) return;

    // 
    att = P->wtm ? wpawn_att [k] : bpawn_att [k];
    att &= P->piece [!P->wtm];

    if (att)
    {
        B->check_is ++;
        if (fast && B->check_is > 1) return;

        B->attacker = bitscan (att);
    };

    // 
    att = knight_att [k] & P->piece[BN + !P->wtm];

    if (att)
    {
        B->check_is ++;
        if (fast && B->check_is > 1) return;

        B->attacker = bitscan (att);
    };

    // 
    t   = P->piece [BK + !P->wtm];
    j   = bitscan (t);
    att = king_att [j];

    if (att & king)
    {
        B->check_is ++;
        B->attacker = j;
    };

    // rnbqkbnr/pp1ppppp/2p5/7Q/8/4P3/PPPP1PPP/RNB1KBNR b KQkq -
    // 18 

    return;
};

//    (  1 )

//   :
// 1.     
// 2.         
int mate_at_a_glance (board * B)
{
    int i, j, k;
    int blc;
    U64 bb, t, ray, occ;
    U64 add, att, p;
    U64 king, safety;
    U64 safatt; //   
    U64 rq, bq;

    king = P->piece[BK + !P->wtm]; //  ,   
    k = bitscan(king);      //  
    safety = king_att[k];  //   

    rq = bq = P->piece [BQ + P->wtm];
    rq |= P->piece [BR + P->wtm];
    bq |= P->piece [BB + P->wtm];

    // ,   
    occ = (P->occupied [0] | P->occupied [1]) & ~king;




    // Step 1.      -   (   )

    // 1.1.    ,      
    att = r_att (P->occupied [!P->wtm], k) & rq;

    // 1.2.          
    for (; att; att ^= t)
    {
        t = LSB (att);
        j = bitscan (t);

        p = link_att [j][k] & P->occupied [P->wtm];

        //  1  
        if (p && !LSB (p))
        {
            //         
            i = bitscan (p);
            blc = P->squares [i] >> 1;

            switch (blc)
            {
                case KNIGHT_ID:

                //   
                //if (n_tropcnt [i][k] == 2)
                {
                };

                break;


            };

        };
    };




    /*

    // +-----------------------------------------------------+
    // |  Condition 1.  ,          |
    // +-----------------------------------------------------+
    ray = r_att (occ, k); //     

    for (bb = P->piece[BR + P->wtm] | P->piece[BQ + P->wtm]; bb; bb ^= t)
    {
        t = LSB (bb);
        i = bitscan (t); //  

        for (att = r_att (occ, i) & ray; att; att ^= p)
        {
            p = LSB (att);
            j = bitscan (p); //  

            if (maag_c2 (i, j, 1)) return 1; //  
        };
    };

    ray = b_att (occ, k); //     

    for (bb = P->piece[BB + P->wtm] | P->piece[BQ + P->wtm]; bb; bb ^= t)
    {
        t = LSB (bb);
        i = bitscan (t);

        for (att = b_att (occ, i) & ray; att; att ^= p)
        {
            p = LSB (att);
            j = bitscan (p);

            if (maag_c2 (i, j, 1)) return 1; //  
        };
    };

    //   ///////////////////////////////////////////////////

    ray = knight_att [k]; //     

    for (bb = P->piece[BN + P->wtm]; bb; bb ^= t)
    {
        t = LSB (bb);
        i = bitscan (t);

        for (att = knight_att [i] & ray; att; att ^= p)
        {
            p = LSB (att);
            j = bitscan (p);

            if (maag_c2 (i, j, 0)) return 1; //  
        };
    };

    //   //////////////////////////////////////////////////

    //    ?



    /*
    if (P->wtm)
    {
        ray = bpawn_att [k] >> 8; //   

        for (bb = P->piece[BP + P->wtm] & (~occ << 8); bb; bb ^= t)
        {
            t = LSB (bb);
            i = bitscan (t);

            if (maag_c2 (i, i + 8, 0)) return 1; //  
        };
    }
    else
    {
        ray = wpawn_att [k] << 8; //   

        att =
    };
    */



    return 0;
};

// Mate at a glance condition 2
int maag_c2 (int fr, int to, int slider, U64 add_att)
{
};

int threefold_repetition (board * B)
{
    int i;
    int count = 0;

    for (i = S->count_moves - 2; i > S->irreversible_ply; i -= 2)
    {
        if (B->hash == R[i].key)
        {
            count ++;
            if (count > 0) return 1;
        };
    };

    return 0;
};

/*// Fruit
static const int KingAttackWeight[16] = {
   0, 0, 128, 192, 224, 240, 248, 252, 254, 255, 256, 256 ,256, 256, 256, 256,
};
*/
