#include "main.h"



void hash_init ()
{
    int i, j;

    srand (111); // was 111

	for (i = 0; i < 12; i ++)
	{
		for (j = 0; j < 64; j ++)
		{
			hash_key [i][j] = rand64();
		};
	};

	for (i = A3; i <= H3; i ++) hash_key_ep [i] = rand64();
	for (i = A6; i <= H6; i ++) hash_key_ep [i] = rand64();

	for (i = 0; i < 16; i ++) hash_key_cast [i] = rand64();

	hash_key_wtm [0] = rand64();

	for (i = 0; i < 2; i ++)
	{
	    for (j = 0; j < 64; j ++)
	    {
	        pawn_key [i][j] = rand64();
	    };
	};

	hash_refresh (H);
	reps_refresh ();
	pent_refresh ();
};

void hash_refresh (hash * _H)
{
    int i;
    hash_t* h;

    for (i = 0; i < _H->size; i ++)
	{
	    h = &_H->entry [i];

        h->age   = 0;
	    h->depth = 0;
	    h->key   = E;
	    h->move  = 0;
	    h->score = 0;
	    h->type  = 0;
	};
};

void reps_refresh ()
{
    int i;

    for (i = 0; i < REP_TABLE; i ++)
	{
	    R[i].key = E;
	    R[i].irreversible_last = 0;
	};
};

void pent_refresh ()
{
    int i;

    for (i = 0; i < PWN_TABLE; i ++)
	{
	    Y[i].key = E;
	    Y[i].score = 0;
	};
};

pentry * pent_get (board * B)
{
    pentry * p;
    int index;

    index = B->phash & (PWN_TABLE - 1);
    p = Y + index;
    if (p->key == B->phash) return p;

    return NULL;
};

void pent_update (int score, eval_t * J, board * B)
{
    pentry * p;
    int index;

    index = B->phash & (PWN_TABLE - 1);
    p = Y + index;

	p->score      = score;
	p->key        = B->phash;
	p->shield[0]  = J->shield[0];
	p->shield[1]  = J->shield[1];
	p->targets[0] = J->targets[0];
	p->targets[1] = J->targets[1];
	p->ofiles[0]  = (uchar) J->ofiles[0];
	p->ofiles[1]  = (uchar) J->ofiles[1];
};

hash_t* hash_get (hash * _H, board * B)
{
    hash_t* h;
    U64 index;

    index = B->hash & (_H->size - 1);
    h = &_H->entry [index];
    if (h->key == B->hash) return h;

    return NULL;
};

U64 hash_calc (board * B)
{
	int i, j;
	U64 bb, t;
	U64 h, p;

	h = p = E;

	for (i = 0; i < 12; i ++)
	{
		bb = P->piece [i];

		while (bb)
		{
			t = LSB (bb);
			j = bitscan (t);

			h ^= hash_key [i][j];
			if (i < 2) p ^= pawn_key [i][j];

			bb ^= t;
		};
	};

	if (P->wtm) h ^= hash_key_wtm [0];

	B->hash_mat = h;
	B->phash = p;

	h ^= hash_key_ep   [B->en_passant];
	h ^= hash_key_cast [B->castling_rights];

	return h;
};

void hash_update (int mv, int score, int type, int depth, hash * _H, board * B)
{
	hash_t* h;
    U64 index;

    index = B->hash & (_H->size - 1);
    h = &_H->entry [index];

	if (h->age == S->g_hash_age && h->depth > depth) return;

	//   ,   
	//if (!mv) mv = h->move; // !     

	if      (score >   MATE) score += S->ply;
	else if (score < - MATE) score -= S->ply;

	h->age   = S->g_hash_age;
	h->depth = depth;
	h->score = score;
	h->key   = B->hash;
	h->move  = mv;
	h->type  = type;
};

void hash_retrieve (int &alpha, int &beta, int &mv, int depth, hash * _H, board * B)
{
    int score;
    hash_t * h = hash_get (_H, B);

	if (h != NULL)
	{
		mv = h->move;

		if (h->depth < depth) return;

        score = h->score;

        if      (score >   MATE) score -= S->ply;
        else if (score < - MATE) score += S->ply;


        // Thanks to CPW
        if (h->type == HASH_EXACT) //  
        {
            alpha = beta = score;
        }
        /*
        else if (h->type == HASH_ALPHA)
		{
			if (score < beta)
			{
				if (score <= alpha) alpha = beta = score;
				else beta = score;
			};
		}
		else if (h->type == HASH_BETA)
		{
			if (score > alpha)
			{
				if (score >= beta) alpha = beta = score;
				else alpha = score;
			};
		};
		*/
        else if (h->type == HASH_ALPHA) //  
        {
            if (score <= alpha) alpha = score;
        }
        else if (h->type == HASH_BETA) //  
        {
            if (score >= beta) beta = score;
        };
	}
	else mv = 0; //   
};
