#include "main.h"

void init_attacks ()
{
	int a, b, c, i, dir;
	int st, en;
	U64 bb, att;

	// Knight attacks
	for (a = 0; a < 8; a ++)
	{
		for (b = 0; b < 8; b ++)
		{
			c = compose (a, b);
			bb = I << c;

			att  = (bb & ~(file_a | file_b)) << 6;
			att |= (bb & ~(file_h | file_g)) << 10;
			att |= (bb & ~file_a) << 15;
			att |= (bb & ~file_h) << 17;

			att |= (bb & ~(file_h | file_g)) >> 6;
			att |= (bb & ~(file_a | file_b)) >> 10;
			att |= (bb & ~file_h) >> 15;
			att |= (bb & ~file_a) >> 17;

			knight_att [c] = att;
		};
	};

	// King attacks
	for (a = 0; a < 8; a ++)
	{
		for (b = 0; b < 8; b ++)
		{
			c = compose (a, b);
			bb = I << c;

			att  = (bb & ~file_h) << 1;
			att |= (bb & ~file_a) << 7;
			att |= bb << 8;
			att |= (bb & ~file_h) << 9;

			att |= (bb & ~file_a) >> 1;
			att |= (bb & ~file_h) >> 7;
			att |= bb >> 8;
			att |= (bb & ~file_a) >> 9;

			king_att [c] = att;
		};
	};

	// Pawn attacks
	for (a = 0; a < 8; a ++)
	{
		for (b = 0; b < 8; b ++)
		{
			c = compose (a, b);
			bb = I << c;

			att  = (bb & ~file_a) << 7;
			att |= (bb & ~file_h) << 9;

			wpawn_att [c] = att;

			att  = (bb & ~file_h) >> 7;
			att |= (bb & ~file_a) >> 9;

			bpawn_att [c] = att;
		};
	};

	// Link attacks
	for (a = 0; a < 64; a ++)
	{
		for (b = 0; b < 64; b ++)
		{
			link_att [a][b] = E;

			//uci_send ("%i %i\n", a, b);

			dir = is_on_line (a, b);

			//     
			if (!dir) continue;

			if (a < b)
			{
				st = a;
				en = b;
			}
			else
			{
				st = b;
				en = a;
				dir = -dir;
			};

			for (i = st; i <= en; i += dir)
			{
				link_att [a][b] |= I << i;
			};

			link_att [a][b] &= ~(I << st);
			link_att [a][b] &= ~(I << en);
		};
	};
};

void init_magics ()
{
	int i, j;

	S->err = 0;

	//   
	if (!GEN_MAGICS && S->f_magics != NULL)
	{
		for (i = 0; i < 64; i ++)
		{
			for (j = 0; j < (1 << 12); j ++)
			{
				rook_att [i][j] = f_read_compact (S->f_magics);//f_read64 (S->f_magics);
			};

			rook_m [i].mask  = f_read_compact (S->f_magics);//f_read64 (S->f_magics);
			rook_m [i].magic = f_read_compact (S->f_magics);//f_read64 (S->f_magics);
		};

		for (i = 0; i < 64; i ++)
		{
			for (j = 0; j < (1 << 9); j ++)
			{
				bishop_att [i][j] = f_read_compact (S->f_magics);//f_read64 (S->f_magics);
			};

			bishop_m [i].mask  = f_read_compact (S->f_magics);//f_read64 (S->f_magics);
			bishop_m [i].magic = f_read_compact (S->f_magics);//f_read64 (S->f_magics);
		};

		//    
		if (S->err)
		{
			fclose (S->f_magics);
			goto exit;
		};

		return;
	};

exit:
	//       
	S->f_magics = fopen (F_MAGICS, "wb");

	// Rook magics
	uci_send ("const U64 r_magics[64] = {\n");
	for (i = 0; i < 64; i ++) init_rook_magics (i);
	uci_send ("};\n\n");

	// Bishop magics
	uci_send ("const U64 b_magics[64] = {\n");
	for (i = 0; i < 64; i ++) init_bishop_magics (i);
	uci_send ("};\n\n");

	fclose (S->f_magics);
};

void init_rook_magics (int c)
{
	const int n = 12; //    
	int i, k;        // , 
	int fail;       //  
	int bits;      //    
	U64 bb, key, mask, magic;
	U64 x64; //  popcount'

	U64 blk [1 << n]; //  
	U64 atc [1 << n]; //  

	U64 used [1 << n]; //    -

	bb = I << c;
	mask = rmask (c);
	bits = pop_count (mask);

	//uci_send ("%i ", bits);

	//  
	for (i = 0; i < (1 << bits); i ++)
	{
		key = index_to_64 (i, bits, mask);

		blk [i] = key;
		atc [i] = ratt (c, key);

		//print_this_att (key);
		//print_this_att (atc [i]);
	};

	//   

	for (k = 0; k < 100000000; k ++)
	{
		fail = 0;
		magic = rand64_fewbits ();

		//if ( pop_count ( (mask * magic) & (U64)0xFF00000000000000) < 6 ) continue;

		for (i = 0; i < (1 << n); i ++) used[i] = E;

		for (i = 0; !fail && (i < (1 << bits)); i ++)
		{
			bb = blk [i];
			bb *= magic;
			bb >>= (64 - n);

			//   
			if (used [bb] == E) used [bb] = atc [i];
			else if (used [bb] != atc [i]) fail = 1;
		};
		if (!fail) break;
	};

	if (fail) uci_send ("Failed!!!\n");
	else uci_send (" 0x%08X%08X, \n", (int)(magic >> 32), (int)magic);

	//  
	for (i = 0; i < (1 << n); i ++)
	{
		rook_att [c][i] = used [i];
		f_write_compact (S->f_magics, used [i]);//fprintf (S->f_magics, "%08X%08X", (int)(used [i] >> 32), (int)used [i]);
	};

	rook_m [c].mask = mask;
	rook_m [c].magic = magic;

	f_write_compact (S->f_magics, mask);//fprintf (S->f_magics, "%08X%08X", (int)(mask >> 32), (int)mask);
	f_write_compact (S->f_magics, magic);//fprintf (S->f_magics, "%08X%08X", (int)(magic >> 32), (int)magic);
};

void init_bishop_magics (int c)
{
	const int n = 9;  //    
	int i, k;        // , 
	int fail;       //  
	int bits;      //    
	U64 bb, key, mask, magic;
	U64 x64; //  popcount'

	U64 blk [1 << n]; //  
	U64 atc [1 << n]; //  

	U64 used [1 << n]; //    -

	bb = I << c;
	mask = bmask (c);
	bits = pop_count (mask);

	//uci_send ("%i ", bits);

	//  
	for (i = 0; i < (1 << bits); i ++)
	{
		key = index_to_64 (i, bits, mask);

		blk [i] = key;
		atc [i] = batt (c, key);

		//print_this_att (key);
		//print_this_att (atc [i]);
	};

	//   

	for (k = 0; k < 100000000; k ++)
	{
		fail = 0;
		magic = rand64_fewbits ();

		//if ( pop_count ( (mask * magic) & (U64)0xFF00000000000000) < 6) continue;

		for (i = 0; i < (1 << n); i ++) used[i] = E;

		for (i = 0; !fail && (i < (1 << bits)); i ++)
		{
			bb = blk [i];
			bb *= magic;
			bb >>= (64 - n);

			//   
			if (used [bb] == E) used [bb] = atc [i];
			else if (used [bb] != atc [i]) fail = 1;
		};
		if (!fail) break;
	};

	if (fail) uci_send ("Failed!!!\n");
	else uci_send (" 0x%08X%08X, \n", (int)(magic >> 32), (int)magic);

	//  
	for (i = 0; i < (1 << n); i ++)
	{
		bishop_att [c][i] = used [i];
		f_write_compact (S->f_magics, used [i]);//fprintf (S->f_magics, "%08X%08X", (int)(used [i] >> 32), (int)used [i]);
	};

	bishop_m [c].mask = mask;
	bishop_m [c].magic = magic;

	f_write_compact (S->f_magics, mask);//fprintf (S->f_magics, "%08X%08X", (int)(mask >> 32), (int)mask);
	f_write_compact (S->f_magics, magic);//fprintf (S->f_magics, "%08X%08X", (int)(magic >> 32), (int)magic);
};

U64 rmask (int sq)
{
	U64 result = E;
	int rk, fl, r, f;

	rk = hori (sq);
	fl = vert (sq);

	for (r = rk + 1; r <= 6; r ++) result |= I << compose(r, fl);
	for (r = rk - 1; r >= 1; r --) result |= I << compose(r, fl);
	for (f = fl + 1; f <= 6; f ++) result |= I << compose(rk, f);
	for (f = fl - 1; f >= 1; f --) result |= I << compose(rk, f);

	return result;
};

U64 bmask (int sq)
{
	U64 result = E;
	int rk, fl, r, f;

	rk = hori (sq);
	fl = vert (sq);

	for (r = rk + 1, f = fl + 1; r <= 6 && f <= 6; r ++, f ++) result |= I << compose(r, f);
	for (r = rk + 1, f = fl - 1; r <= 6 && f >= 1; r ++, f --) result |= I << compose(r, f);
	for (r = rk - 1, f = fl + 1; r >= 1 && f <= 6; r --, f ++) result |= I << compose(r, f);
	for (r = rk - 1, f = fl - 1; r >= 1 && f >= 1; r --, f --) result |= I << compose(r, f);

	return result;
};

U64 qmask (int sq)
{
	U64 result = E;
	int rk, fl, r, f;

	rk = hori (sq);
	fl = vert (sq);

	for (r = rk + 1; r <= 7; r ++) result |= I << compose(r, fl);
	for (r = rk - 1; r >= 0; r --) result |= I << compose(r, fl);
	for (f = fl + 1; f <= 7; f ++) result |= I << compose(rk, f);
	for (f = fl - 1; f >= 0; f --) result |= I << compose(rk, f);

	for (r = rk + 1, f = fl + 1; r <= 7 && f <= 7; r ++, f ++) result |= I << compose(r, f);
	for (r = rk + 1, f = fl - 1; r <= 7 && f >= 0; r ++, f --) result |= I << compose(r, f);
	for (r = rk - 1, f = fl + 1; r >= 0 && f <= 7; r --, f ++) result |= I << compose(r, f);
	for (r = rk - 1, f = fl - 1; r >= 0 && f >= 0; r --, f --) result |= I << compose(r, f);

	return result;
};

U64 ratt (int sq, U64 block)
{
	U64 bb, result = E;
	int rk, fl, r, f;

	rk = hori (sq);
	fl = vert (sq);

	for (r = rk + 1; r <= 7; r ++)
	{
		bb = I << compose(r, fl);
		result |= bb;
		if (block & bb) break;
	};
	for (r = rk - 1; r >= 0; r --)
	{
		bb = I << compose(r, fl);
		result |= bb;
		if (block & bb) break;
	};
	for (f = fl + 1; f <= 7; f ++)
	{
		bb = I << compose(rk, f);
		result |= bb;
		if (block & bb) break;
	};
	for (f = fl - 1; f >= 0; f --)
	{
		bb = I << compose(rk, f);
		result |= bb;
		if (block & bb) break;
	};

	return result;
};

U64 batt (int sq, U64 block)
{
	U64 bb, result = E;
	int rk, fl, r, f;

	rk = hori (sq);
	fl = vert (sq);

	for (r = rk + 1, f = fl + 1; r <= 7 && f <= 7; r ++, f ++)
	{
		bb = I << compose(r, f);
		result |= bb;
		if (block & bb) break;
	};
	for (r = rk + 1, f = fl - 1; r <= 7 && f >= 0; r ++, f --)
	{
		bb = I << compose(r, f);
		result |= bb;
		if (block & bb) break;
	};
	for (r = rk - 1, f = fl + 1; r >= 0 && f <= 7; r --, f ++)
	{
		bb = I << compose(r, f);
		result |= bb;
		if (block & bb) break;
	};
	for (r = rk - 1, f = fl - 1; r >= 0 && f >= 0; r --, f --)
	{
		bb = I << compose(r, f);
		result |= bb;
		if (block & bb) break;
	};

	return result;
};

U64 index_to_64 (int index, int bits, U64 m)
{
	int i;
	U64 t, result = E;

	for (i = 0; i < bits; i++)
	{
		t = LSB (m);
		if (index & (1 << i)) result |= t;
		m ^= t;
	};

	return result;
};

