// +----------------------------------------------------------------------------------+
// |      main.h - header file, written by Nikita Kurgin (aka neurocore), 2012        |
// +----------------------------------------------------------------------------------+

#include <stdio.h>
#include <stdlib.h>
#include <cstring>
#include <ctime>
#include <cstdarg>
//#include <windows.h>
//#include <conio.h>
#include <math.h>

#ifndef _WIN32
#define _WIN32
#endif

// JA corrections
#if defined(_WIN32) || defined(_WIN64)
#include <windows.h>
#include <conio.h>
#else
#include <unistd.h>
#endif

#define LOGGING
#undef LOGGING

#define AUTOTUNE
#undef AUTOTUNE

#ifndef AUTOTUNE
    #define LAZY_EVAL
    //#undef LAZY_EVAL

    #define PV_OUTPUT
    //#undef PV_OUTPUT

//#endif
    #define UCI_OUT
    //#undef UCI_OUT
#endif

#define PAWN_HASH
//#undef PAWN_HASH

#define USE_FAST_XRAYS
//#undef USE_FAST_XRAYS

#define SOFT_CONTROL
//#undef SOFT_CONTROL

#define USE_FUTILITY // deleted
#undef USE_FUTILITY

#define DEBUG_TUNING
#undef DEBUG_TUNING

#define DEBUG_EVAL
#undef DEBUG_EVAL


#define C64(a) (a##ULL)

typedef   signed long long  S64;
typedef unsigned long long  U64;
typedef unsigned int        U32;
typedef unsigned char     uchar;

#define var int

#define NAME      "Eia"
#define VERS      "0.2" // Don't specify here

// For 32/64 bit compilation
#if defined(_WIN64)
#define VERSION   VERS" w64"
#else
#if defined(_WIN32)
#define VERSION   VERS" w32"
#endif
#endif

// Linux detection
#if !defined(_WIN32) && !defined(_WIN64)
#if defined(__i386__)
# define IS32BIT
#else
# define IS64BIT
#endif

#if defined(IS32BIT)
#define VERSION   VERS" lin32"
#endif
#if defined(IS64BIT)
#define VERSION   VERS" lin64"
#endif
#endif

#define AUTHOR    "Nikita Kurgin"

#define F_REPORT  "report.log"
#define F_MAGICS  "magics.bin"
#define F_CONFIG  "default.var"

#define USC_WIN_SCORE     300             // ,    (   )
#define USC_TIME          3              //     
#define USC_POINTS        32            //      
#define USC_TESTS         32           //  -   
#define USC_DEPTH         3           //       
#define USC_RADIUS        EINT(10)   //      
#define USC_GAMMA         EINT(10)  // ,      
#define USC_DICHOT        3        //         ()
#define USC_DELTA         2       //      
#define EVAL_MUL          128
#define MAX_DEPTH         100
#define GO_MAX_DEPTH      98
#define BENCH_LW_DEPTH    1
#define BENCH_HG_DEPTH    6
#define ROOT_ORDERING     2                 //      
#define SINGULAR_VAL      400              //      
#define DEPTH_TO_MOVE     4               //  ,    
#define SUDDEN_DEATH_TIME 1              // ,      ()
#define USE_NULL_MOVE     1             //    ?
#define HASH_SIZE         (1 << 20)    //     (  )
#define PV_SIZE           (1 << 16)   //     
#define REP_TABLE         5900       //   
#define PWN_TABLE         (1 << 18) //   
#define CHECK_PLY_LIMIT   7        //    
#define PPUSH_PLY_LIMIT   6       //    
#define NODES_OF_SILENCE  (1 << 24)
#define NODES_OF_INPUT    (1 << 18)
#define DEPTH_TO_CURRMV   9   //     currmove  
#define GEN_MAGICS        0  //     ?
#define RAZOR_DEPTH       4 // Maximum depth for razoring

//     (LERF-mapping)

#define E         C64(0x0000000000000000)
#define I         C64(0x0000000000000001)
#define U         C64(0xFFFFFFFFFFFFFFFF)

#define file_a    C64(0x0101010101010101)
#define file_b    C64(0x0202020202020202)
#define file_c    C64(0x0404040404040404)
#define file_d    C64(0x0808080808080808)
#define file_e    C64(0x1010101010101010)
#define file_f    C64(0x2020202020202020)
#define file_g    C64(0x4040404040404040)
#define file_h    C64(0x8080808080808080)

#define rank_1    C64(0x00000000000000FF)
#define rank_2    C64(0x000000000000FF00)
#define rank_3    C64(0x0000000000FF0000)
#define rank_4    C64(0x00000000FF000000)
#define rank_5    C64(0x000000FF00000000)
#define rank_6    C64(0x0000FF0000000000)
#define rank_7    C64(0x00FF000000000000)
#define rank_8    C64(0xFF00000000000000)

#define debruijn  C64(0x07EDD5E59A4E28C2)

const int index_bsc [64]  =  {63,  0, 58,  1, 59, 47, 53,  2,
                              60, 39, 48, 27, 54, 33, 42,  3,
                              61, 51, 37, 40, 49, 18, 28, 20,
                              55, 30, 34, 11, 43, 14, 22,  4,
                              62, 57, 46, 52, 38, 26, 32, 41,
                              50, 36, 17, 19, 29, 10, 13, 21,
                              56, 45, 25, 31, 35, 16,  9, 12,
                              44, 24, 15,  8, 23,  7,  6,  5};

//  , 

enum    {BP, WP,
		 BN, WN,
		 BB, WB,
		 BR, WR,
		 BQ, WQ,
		 BK, WK,
		 no_p = 255};

enum    {A1, B1, C1, D1, E1, F1, G1, H1,
	 	 A2, B2, C2, D2, E2, F2, G2, H2,
	 	 A3, B3, C3, D3, E3, F3, G3, H3,
	 	 A4, B4, C4, D4, E4, F4, G4, H4,
	 	 A5, B5, C5, D5, E5, F5, G5, H5,
	 	 A6, B6, C6, D6, E6, F6, G6, H6,
	 	 A7, B7, C7, D7, E7, F7, G7, H7,
	 	 A8, B8, C8, D8, E8, F8, G8, H8};


enum    {ST_EXIT, ST_ANALYSING, ST_WAITING}; //    (state)
enum    {HASH_EXACT, HASH_ALPHA, HASH_BETA}; // 

#define START_FEN "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"


//  (__|___)

enum   {f_none     = 0,
        f_capture  = 4,
        f_castling = 8,
        f_prom     = 16,
        f_ep       = 32,
        f_check    = 64,
        f_2check   = 128,

        f_mask     = 252, // 111111|00

        f_color    = 1,
        BLACK      = 0,
        WHITE      = 1};

//    

enum   {o_pv        =  0x08000000,
        o_evasion   =  0x04000000,
        o_hash      =  0x02000000,
        o_recapture =  0x01000000,
        o_prom      =  0x00800000,
        o_good_cap  =  0x00400000,
        o_bad_cap   =  0x00200000,
        o_killer    =  0x00100000,
        o_quiet     =  0x00000000,

//        o_tactical  =  0x00000010,
//        o_pawn_4    =  0x00100000,
//        o_pawn_5    =  0x00200000,
//        o_pawn_6    =  0x00400000
        };

//   (   )

#define w_p   100
#define w_b   325
#define w_n   325
#define w_r   500
#define w_q   975
#define w_k   10000

#define b_p   100
#define b_b   325
#define b_n   325
#define b_r   500
#define b_q   975
#define b_k   10000

//  P < N <= B < R < Q < K

#define DRAW  0     //  (   !)
#define WIN   10000 //  
#define MATE  32000 //  
#define INF   32767

#define _k1  C64(0x5555555555555555)
#define _k2  C64(0x3333333333333333)
#define _k4  C64(0x0f0f0f0f0f0f0f0f)
#define _kf  C64(0x0101010101010101)

//   

//     (thanks to GULL)
#define MAT_WQ 1
#define MAT_BQ 3
#define MAT_WR 3 * 3
#define MAT_BR 3 * 3 * 3
#define MAT_WB 3 * 3 * 3 * 3
#define MAT_BB 3 * 3 * 3 * 3 * 3
#define MAT_WN 3 * 3 * 3 * 3 * 3 * 3
#define MAT_BN 3 * 3 * 3 * 3 * 3 * 3 * 3
#define MAT_WP 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3
#define MAT_BP 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 9
#define TOTAL_MAT (2*(MAT_WQ + MAT_BQ + MAT_WR + MAT_BR + MAT_WB + MAT_BB + MAT_WN + MAT_BN)+ \
                   8*(MAT_WP + MAT_BP) + 1)
#define UNUSIAL_MAT (1 << 20)

//  
#define EVAL_USUAL     0
#define EVAL_KPK       1
#define EVAL_KBBK      2
#define EVAL_KBNK      3
#define EVAL_KQKP      4
#define EVAL_KRKP      5
#define EVAL_KQKR      6
#define EVAL_KRKN      7
#define EVAL_BISHOPS   8

//  
#define STAGE_EG    9  //      

//   
#define _A1  (I << A1)
#define _B1  (I << B1)
#define _C1  (I << C1)
#define _D1  (I << D1)
#define _E1  (I << E1)
#define _F1  (I << F1)
#define _G1  (I << G1)
#define _H1  (I << H1)

#define _A2  (I << A2)
#define _B2  (I << B2)
#define _C2  (I << C2)
#define _D2  (I << D2)
#define _E2  (I << E2)
#define _F2  (I << F2)
#define _G2  (I << G2)
#define _H2  (I << H2)

#define _A3  (I << A3)
#define _B3  (I << B3)
#define _C3  (I << C3)
#define _D3  (I << D3)
#define _E3  (I << E3)
#define _F3  (I << F3)
#define _G3  (I << G3)
#define _H3  (I << H3)

#define _A4  (I << A4)
#define _B4  (I << B4)
#define _C4  (I << C4)
#define _D4  (I << D4)
#define _E4  (I << E4)
#define _F4  (I << F4)
#define _G4  (I << G4)
#define _H4  (I << H4)

#define _A5  (I << A5)
#define _B5  (I << B5)
#define _C5  (I << C5)
#define _D5  (I << D5)
#define _E5  (I << E5)
#define _F5  (I << F5)
#define _G5  (I << G5)
#define _H5  (I << H5)

#define _A6  (I << A6)
#define _B6  (I << B6)
#define _C6  (I << C6)
#define _D6  (I << D6)
#define _E6  (I << E6)
#define _F6  (I << F6)
#define _G6  (I << G6)
#define _H6  (I << H6)

#define _A7  (I << A7)
#define _B7  (I << B7)
#define _C7  (I << C7)
#define _D7  (I << D7)
#define _E7  (I << E7)
#define _F7  (I << F7)
#define _G7  (I << G7)
#define _H7  (I << H7)

#define _A8  (I << A8)
#define _B8  (I << B8)
#define _C8  (I << C8)
#define _D8  (I << D8)
#define _E8  (I << E8)
#define _F8  (I << F8)
#define _G8  (I << G8)
#define _H8  (I << H8)

//        
#define DIST_1   (1 << 0)
#define DIST_2   (1 << 1)
#define DIST_3   (1 << 2)
#define DIST_4   (1 << 3)

//   
#define MATE_A2  (1 << 8)
#define MATE_C2  (1 << 9)
#define MATE_B1  (1 << 10)
#define MATE_B3  (1 << 11)
#define MATE_A1  (1 << 12)
#define MATE_A3  (1 << 13)
#define MATE_C1  (1 << 14)
#define MATE_C3  (1 << 15)

//    
#define BLACK_COLOR   C64(0xAA55AA55AA55AA55)
#define WHITE_COLOR   C64(0x55AA55AA55AA55AA)

// 
#define CENTER        (_E4 | _E5 | _D4 | _D5 )

#define W_BIG_CENTER  (_C4 |             _F4 | \
                       _C5 |             _F5 | \
                       _C6 | _D6 | _E6 | _F6 )

#define B_BIG_CENTER  (_C3 | _D3 | _E3 | _F3 | \
                       _C4 |             _F4 | \
                       _C5 |             _F5 )

//  
#define W_SHIELD_MASK (_A2 | _A3 | _A4 | _A5 | \
                       _B2 | _B3 | _B4 | _B5 | \
                       _C2 | _C3 | _C4 | _C5 | \
                                               \
                       _F2 | _F3 | _F4 | _F5 | \
                       _G2 | _G3 | _G4 | _G5 | \
                       _H2 | _H3 | _H4 | _H5 )

#define B_SHIELD_MASK (_A7 | _A6 | _A5 | _A4 | \
                       _B7 | _B6 | _B5 | _B4 | \
                       _C7 | _C6 | _C5 | _C4 | \
                                               \
                       _F7 | _F6 | _F5 | _F4 | \
                       _G7 | _G6 | _G5 | _G4 | \
                       _H7 | _H6 | _H5 | _H4 )

//  
#define Q_SIDE     (file_a | file_b | file_c | file_d)
#define K_SIDE     (file_e | file_f | file_g | file_h)

#define W_HALF     (rank_1 | rank_2 | rank_3 | rank_4)
#define B_HALF     (rank_5 | rank_6 | rank_7 | rank_8)

// 
#define W_OUTPOST (((file_c | file_d | file_e | file_f) ^ rank_8) & B_HALF)
#define B_OUTPOST (((file_c | file_d | file_e | file_f) ^ rank_1) & W_HALF)


//  up_mask  dn_mask
#define UPMASK (U ^ rank_1)
#define DNMASK (U ^ rank_8)

//  
#define PAWN_ID      0
#define KNIGHT_ID    1
#define BISHOP_ID    2
#define ROOK_ID      3
#define QUEEN_ID     4
#define KING_ID      5
#define NOPIECE_ID   7

//   
#define c_K    1
#define c_Q    2
#define c_k    4
#define c_q    8

//      
#define CTRL_BP      (1 << 4)
#define CTRL_BN1     (1 << 5)
#define CTRL_BN2     (1 << 6)
#define CTRL_BB      (1 << 7)
#define CTRL_BR1     (1 << 12)
#define CTRL_BR2     (1 << 13)
#define CTRL_BQ      (1 << 14)
#define CTRL_BK      (1 << 15)

#define CTRL_WP      (1 << 20)
#define CTRL_WN1     (1 << 21)
#define CTRL_WN2     (1 << 22)
#define CTRL_WB      (1 << 23)
#define CTRL_WR1     (1 << 28)
#define CTRL_WR2     (1 << 29)
#define CTRL_WQ      (1 << 30)
#define CTRL_WK      (1 << 31)


//   
#define c_att_K (_E1 | _F1 | _G1)
#define c_att_Q (_E1 | _D1 | _C1)
#define c_att_k (_E8 | _F8 | _G8)
#define c_att_q (_E8 | _D8 | _C8)

//   
#define c_mov_K (_F1 | _G1)
#define c_mov_Q (_D1 | _C1 | _B1)
#define c_mov_k (_F8 | _G8)
#define c_mov_q (_D8 | _C8 | _B8)



//     ////////////////////////////////////////////////////////////////

#define UCI_SEND_VALUE(a) (uci_send(#a": %i\n", V->v[a]))

#define EINT(a)           ((a) * EVAL_MUL) // Extended int
#define LINT(a)           ((a) / EVAL_MUL) // Lower int

// 
#define TEMPO             (T->v[_TEMPO])

//  
#define HOLES             (T->v[_HOLES])
#define BACKWARD          (T->v[_BACKWARD])
#define BLOCKED_PAWN      (T->v[_BLOCKED_PAWN])
#define ISOLATED          (T->v[_ISOLATED])

//  
#define N_OUTPOST         (T->v[_N_OUTPOST])
#define COLOR_WEAK        (T->v[_COLOR_WEAK]) // TODO
#define FIANCHETTO        (T->v[_FIANCHETTO])
#define BISHOP_PAIR       (T->v[_BISHOP_PAIR])
#define RET_BISHOP        (T->v[_RET_BISHOP])
#define BAD_BISHOP        (T->v[_BAD_BISHOP])
#define TRAPPED_BISHOP    (T->v[_TRAPPED_BISHOP])
#define OPP_BISHOPS       (T->v[_OPP_BISHOPS])
#define TRAPPED_ROOK      (T->v[_TRAPPED_ROOK])
#define BAD_ROOK          (T->v[_BAD_ROOK])
#define TARRASCH_RULE     (T->v[_TARRASCH_RULE])
#define LINKED_ROOKS      (T->v[_LINKED_ROOKS])
#define ROOK_ON_OPENED    (T->v[_ROOK_ON_OPENED])
#define ROOK_ON_SEMI      (T->v[_ROOK_ON_SEMI])
#define ROOK_ON_7TH       (T->v[_ROOK_ON_7TH]) //     7      8 

//  
#define CONTACT_CHECK     (T->v[_CONTACT_CHECK])
#define FORK              (T->v[_FORK])

//  
#define CLOSED_KING       (T->v[_CLOSED_KING])

//  
#define DISC_ATT          (T->v[_DISC_ATT])
#define DISC_CHECK        (T->v[_DISC_CHECK])
#define PIN_PART          (T->v[_PIN_PART])
#define PIN               (T->v[_PIN])
#define PIN_ABS           (T->v[_PIN_ABS])
#define SKEWER            (T->v[_SKEWER])
#define XRAY              (T->v[_XRAY])

//  
#define CONTROL_P         (T->v[_CONTROL_P])
#define CONTROL_N         (T->v[_CONTROL_N])
#define CONTROL_B         (T->v[_CONTROL_B])
#define CONTROL_R         (T->v[_CONTROL_R])
#define CONTROL_Q         (T->v[_CONTROL_Q])
#define CONTROL_K         (T->v[_CONTROL_K])

// 
#define PST_MUL(a)        (T->v[_PST_MUL + a]) // 6
#define KING_DEV(a)       (T->v[_KING_DEV + a]) // 7

//    , , 
#define SYNERGY(a)        (T->v[_SYNERGY + a]) // 8
#define MCENTER(a)        (T->v[_MCENTER + a]) // 5
#define DOUBLED(a)        (T->v[_DOUBLED + a]) // 8

//    
#define SHIELD(a)         (T->v[_SHIELD + a]) // 12
#define STORM(a)          (T->v[_STORM + a])  // 12

//   
#define IMMOBILITY(a)     (T->v[_IMMOBILITY + a]) // 3

// 
#define PASSER(a)         (T->v[_PASSER + a])     // 7
#define FRE_PASSER(a)     (T->v[_FRE_PASSER + a]) // 7
#define DEF_PASSER(a)     (T->v[_DEF_PASSER + a]) // 7

//          (   )
#define NTROP(a)          (T->v[_NTROP + a]) // 6
#define BTROP(a)          (T->v[_BTROP + a]) // 3
#define RTROP(a)          (T->v[_RTROP + a]) // 3
#define QTROP(a)          (T->v[_QTROP + a]) // 3

//  ////////////////////////////////////////////////////////////////////////////
#define _TEMPO             0

//  
#define _HOLES             1
#define _BACKWARD          2
#define _BLOCKED_PAWN      3
#define _ISOLATED          4

//  
#define _N_OUTPOST         5
#define _COLOR_WEAK        6 // TODO
#define _FIANCHETTO        7
#define _BISHOP_PAIR       8
#define _RET_BISHOP        9
#define _BAD_BISHOP        10
#define _TRAPPED_BISHOP    11 // TODO
#define _OPP_BISHOPS       12
#define _BAD_ROOK          13
#define _TRAPPED_ROOK      14 // TODO
#define _TARRASCH_RULE     15
#define _LINKED_ROOKS      16
#define _ROOK_ON_OPENED    17
#define _ROOK_ON_SEMI      18
#define _ROOK_ON_7TH       19 //     7      8 

//  
#define _CONTACT_CHECK     20
#define _FORK              21 // TODO

//  
#define _CLOSED_KING       22

//  
#define _DISC_ATT          23
#define _DISC_CHECK        24
#define _PIN_PART          25
#define _PIN               26
#define _PIN_ABS           27
#define _SKEWER            28
#define _XRAY              29

//  
#define _CONTROL_P         30
#define _CONTROL_N         31
#define _CONTROL_B         32
#define _CONTROL_R         33
#define _CONTROL_Q         34
#define _CONTROL_K         35

// 
#define _PST_MUL           36 // 6
#define _KING_DEV          42 // 7

//    , , 
#define _SYNERGY           49 // 8
#define _MCENTER           57 // 5
#define _DOUBLED           62 // 8

//    
#define _SHIELD            70 // 12
#define _STORM             82 // 12

//   
#define _IMMOBILITY        94 // 3

// 
#define _PASSER            96  // 7 -   !    6 
#define _FRE_PASSER        102 // 7 -   !    6 
#define _DEF_PASSER        108 // 7 -   !    6 

//          (   )
#define _NTROP             114 // 6 -   !
#define _BTROP             119 // 3 -   !
#define _RTROP             121 // 3 -   !
#define _QTROP             123 // 3 -   !


//  
#define CONSTS_N           125


////////////////////////////////////////////////////////////////////////////////////////////




//  ///////////////////////////////////////////////////////////////////////////////////////

#define hori(a)           ( (a) >> 3 )
#define vert(a)           ( (a) & 7 )
#define compose(a, b)     ( ((a) << 3) | b )
#define min(a, b)         ( (a) < (b) ? (a) : (b) )
#define max(a, b)         ( (a) > (b) ? (a) : (b) )

//  
#define absu(a, b)        ( (a) > (b) ? (a) - (b) : (b) - (a) )

//  
#define is_capture(a)     ( (a) & f_capture  )
#define is_castling(a)    ( (a) & f_castling )
#define is_prom(a)        ( (a) & f_prom     )
#define is_ep(a)          ( (a) & f_ep       )
#define is_check(a)       ( (a) & f_check    )
#define is_2check(a)      ( (a) & f_2check   )

#define is_cap_or_prom(a) ( (a) & (f_capture | f_prom) )

//    
#define is_linr(a)        ( ((a - 2) & 12) == 4 ) // 01xx
#define is_diag(a)        ( ((a - 2) & 10) == 2 ) // 0x1x

//    (0 - , 1 - , 2 - , 3 - , 4 - , 5 - , 7 - )
#define piece_id(a)       (   (a) >> 1 )
#define piece_col(a)      (   (a) & 1  )
#define is_white(a)       (  ((a) & 1) )
#define is_black(a)       ( !((a) & 1) )
#define to_black(a)       ( (a) & 0xFFFFFFFE )

//  
#define random(a)         ( rand ()%(a) )
#define is_mate(a)        ( (a) < -MATE || (a) > MATE )
#define exc_draw(a)       ( (a) == DRAW ? (a) + 1 : (a) )

//   ( )
#define val_att_king(a)   ( lo_king_att [a & 0xFF][(a >> 16) & 0xFF] ? lo_king_att [a & 0xFF][(a >> 16) & 0xFF] : hi_king_att [(a >> 8) & 0xFF][a >> 24])
#define val_att_weak(a)   ( lo_weak_att [a & 0xFF][(a >> 16) & 0xFF] ? lo_weak_att [a & 0xFF][(a >> 16) & 0xFF] : hi_weak_att [(a >> 8) & 0xFF][a >> 24])

//    
#define LSB(a)            ( ((a) & (E - (a))) )
#define RLSB(a)           ( ((a) & ((a) - 1)) )

// Cut Shift Left (Right) -     ( ). 
//                               ( )
//                               0  7 !
#define csl(a, sh)        ( ( (a) & shifts_cut_l[sh] ) >> sh )
#define csr(a, sh)        ( ( (a) & shifts_cut_r[sh] ) << sh )

//     (  - LSB)
#define bitscan(a)        ( index_bsc[( LSB(a) * debruijn ) >> 58] )

//   , , 
#define r_att(a, sq)      (  rook_att   [sq][( ( (a) &  rook_m[sq].mask  ) *  rook_m[sq].magic  ) >> (64 - 12)] )
#define b_att(a, sq)      (  bishop_att [sq][( ( (a) & bishop_m[sq].mask ) * bishop_m[sq].magic ) >> (64 -  9)] )
#define q_att(a, sq)      ( r_att(a, sq) | b_att(a, sq) )

// 
#define PIE(a)            ( ((a) >> 24) & 0xFF )
#define FROM(a)           ( ((a) >> 16) & 0xFF )
#define TO(a)             ( ((a) >>  8) & 0xFF )
#define FLAGS(a)          ( (a)         & 0xFF )

//      ( x64!)
#define fill_up(a) (x64  = (U64)(a),     \
                    x64 |= (x64 << 32),  \
                    x64 |= (x64 << 16),  \
                    x64 |  (x64 << 8)    )

#define fill_dn(a) (x64  = (U64)(a),     \
                    x64 |= (x64 >> 32),  \
                    x64 |= (x64 >> 16),  \
                    x64 |  (x64 >> 8)    )

//  
#define generate_moves(a, b, c, d, B)   (B->moves [B->size ++] = ((a) << 24) | ((b) << 16) | ((c) << 8) | (d))
#define generate_proms(a, b, c, d, B, m) ((m) = ((a) << 24) | ((b) << 16) | ((c) << 8) | ((d) + f_prom + 0), \
                                          B->moves [B->size ++] = (m) ++, \
                                          B->moves [B->size ++] = (m) ++, \
                                          B->moves [B->size ++] = (m) ++, \
                                          B->moves [B->size ++] = (m) ++  )

// 
#define move_is_killer(a,b)             (((a) == b->killer1) || ((a) == b->killer2))

// Population Count
#define pop_count(x)   (    x64 = (U64)(x),                          \
                            x64 =  x64        - ((x64 >> 1)  & _k1), \
                            x64 = (x64 & _k2) + ((x64 >> 2)  & _k2), \
                            x64 = (x64        +  (x64 >> 4)) & _k4 , \
                            (int) ((x64 * _kf) >> 56)                )

#define pop_count_few(x) pop_count(x)

//  (     )
/*
#define r_xray(a, sq)  (  r_att (a ^ (r_att(a, sq) & a), sq) & a  )
#define b_xray(a, sq)  (  b_att (a ^ (b_att(a, sq) & a), sq) & a  )
#define q_xray(a, sq)  (  q_att (a ^ (q_att(a, sq) & a), sq) & a  )
*/

//    (, )
#define dist(a, b)     (  absu(hori(a), hori(b)) + absu(vert(a), vert(b))  )
#define manh(a, b)     (  max ( absu(hori(a), hori(b)) , absu(vert(a), vert(b)) ) )

//       ?
#define mat_for_null   (P->piece[BR + P->wtm] ||                                 \
                        P->piece[BQ + P->wtm] ||                                 \
                        pop_count (P->piece[BN + P->wtm] | P->piece[BB + P->wtm]) > 1  )

//     ( -  )
//#define up_mask(a)     ( UPMASK << ((hori(a)) << 3) )
//#define dn_mask(a)     ( DNMASK >> ((7 - (hori(a))) << 3) )

//     
//#define expand(a)      ( a |= a << 8, a |= a << 16, a |= a << 32 )

//  16- 
#define print_hex(a)  ( uci_send ("0x%08X%08X, \n", (int)((a) >> 32), (int)(a)) )
#define print_hex64(a) ( uci_send ("C64(0x%08X%08X), ", (int)((a) >> 32), (int)(a)) )


#define reduction(d, mn, _pv)  ( reductions[_pv][min(d / 2, 63)][min(mn, 63)] )

//  /////////////////////////////////////////////////////////////////////////////////////

struct magic_t
{
	U64 mask;
	U64 magic;
};

#pragma pack(push)
#pragma pack(1)

struct hash_t
{
	U64 key;
	int age;
	int move;
	char type;
	char depth;
	short score;
};

struct pentry //      ,  
{
	U64 key;
	int score;
	int shield[2];
	U64 targets[2];
	uchar ofiles[2];
};

struct mat_t
{
    short score;  // 
    uchar type;    //  
    uchar overload; //    
    uchar stage[2];  //     
};

#pragma pack (pop)

struct reps_t // 
{
    U64 key;
    int irreversible_last; //        
};

struct pv_t
{
    U64 key;
    U64 key_new;
    int mv;
};

struct search;
struct board;
struct eval_t;
struct pos_t;
struct tune_t;

class hash;



//////////////////////////////////////////////////////////////////////////////////////////////////

class hash
{

public:

    hash_t *entry;
    int size;

    hash (int sz = HASH_SIZE)
    {
        U64 x64;

        sz /= sizeof (hash_t);
        int k = pop_count (sz); //      

        for (int i = 1; i < k; i ++) sz ^= LSB(sz);

        entry = new hash_t [sz];
        size = sz;

        //printf ("info string hash size is %i Mb now\n", (sz * sizeof (hash_t)) >> 20);

        for (int i = 0; i < sz; i ++)
        {
            entry [i].key = E;
            entry [i].age = 0;
            entry [i].move = 0;
            entry [i].type = 0;
            entry [i].depth = 0;
            entry [i].score = 0;
        };
    };

    ~hash ()
    {
        delete [] entry;
    };
};

extern hash * H;     // 
extern pentry * Y;    //  
extern board * _B;     //  ()
extern eval_t * J;      // 
extern search * S;       //  
extern tune_t * T0, * T;  //  ,  ()
extern reps_t * R;      //  
extern pos_t * P;     //  ()


//////////////////////////////////////////////////////////////////////////////////////////////////

struct mate_t
{
    int beyond;     //    (AAAAAAAA|BBBBBBBB,  A - , B - . )
    int disc_nq;   //   
    int disc_rbq; //      (  )
    int close;   // 
};

//////////////////////////////////////////////////////////////////////////////////////////////////

struct eval_t
{
    int bad_develop[2]; //   
    U64 pawn_att[2];   //  

    U64 rq[2];  //  
    U64 bq[2]; //   

    int king_stage[2];   //   
    U64 king_safety[2];   //   
    int king_sq[2];

    int shield[2];  // Shield & Storm
    U64 passers[2];  // 

    int q_is[2];    //  
    U64 targets[2];  //  , 
    U64 ofiles[2];    //   
    int stage[2];       //   (0..16)

    int mob[2];
    int atts[64];    //  /   
    U64 no_atts;      //  ,    
    U64 att_def;       //  ,     ( , , )

    U64 occ, empty;
    U64 not_black;
    U64 not_white;
};

//////////////////////////////////////////////////////////////////////////////////////////////////

struct pos_t
{
    int wtm;               //  ?
    U64 piece [12];       //   
    U64 occupied [2];    // 
    int squares [64];   //   
};

//////////////////////////////////////////////////////////////////////////////////////////////////

struct board
{
    int check_is;           // - 
	int attacker;          //   
	int castling_rights;  //    (0000QKqk)
	int en_passant;      // ,      

	U64 hash;
	U64 hash_mat;        //       ( )
	U64 phash;           //  -

	int mat_key;         //   

	int taken;           //  

    U64 msk_bb;          // Mask BB

    int bmove;           //  
    int cmove;          //  
    int hmove;         //   

    int skip_null;      //       ?
    int red;           //   (reduction)
    int n_evasions;   //    ,  
    int killer1;     // 
    int killer2;     //  

    int stage;        //   
    int curr;        //   
	int size;
	int moves [239];
};


//////////////////////////////////////////////////////////////////////////////////////////////////

struct search
{
    int g_hash_age;       // Aging
    board * B_start;      //     

    FILE* f_out;
    FILE* f_magics;
    FILE* f_var;
    FILE* f_pgn;

    int err;              //  

    int state;            // 
    int infinite;         //  
    int score;            //   
    int score_last;       //   
    int best_move;        //   
    int draw;             //    (  )

    int ready_to_search;  //    (  )
    int deferred_cmd_go;  //   go

    clock_t curr_time;

    int think_time;       //    
    int elaps_time;       //    
    int critl_time;       //  

    int is_critical;      //  ? ( ?)
    int sudden_death;     //    !

    int pv_len;           //    
    U64 pv_start;         //     pv  

    U64 nodes_count;      //    ()
    U64 nodes;            //  ,   

    int ply;
    int search_depth;     //   

    int irreversible_ply; //      
    int fifty;            //    50- 
	int count_moves;      //  

	char mv_out [8];      //  

	int tuning;           //    
	int tune_score;       //      
	int games;            //   
};

////////////////////////////////////////////////////////////////////////////////////////////////


//     ,    
struct tune_t
{
    var v[CONSTS_N];
};


//////    () /////////////////////////////////////////////////////////

//  
extern U64 hash_key [13][64];  // 
extern U64 hash_key_ep [64];   //   
extern U64 hash_key_cast [16]; //   
extern U64 hash_key_wtm [1];   //  

extern U64 pawn_key [2][64];   // 

extern const int  tune_penalty [255];
extern const int  mat_keys     [12]; //     
extern const int  material     [12];
extern const char piece_array  [13];

//     
extern U64 w_att_frontspan [64];
extern U64 b_att_frontspan [64];
extern U64 w_wide_frontspan [64];
extern U64 b_wide_frontspan [64];

//        
extern U64 bad_rook [64];

//   
extern U64 expand [256];
extern U64 up_mask [64];
extern U64 dn_mask [64];

//  
extern const int wshield_sq    [64];
extern const int bshield_sq    [64];

//    
extern int king_sfty_sq [64*8];

//   
extern const U64 ranks         [8];
extern const U64 files         [8];

//      ()
extern const U64 lines         [64];

//    
extern const U64 diags         [64];

//       
extern const U64 shifts_cut_l  [8];
extern const U64 shifts_cut_r  [8];

//   
extern mat_t mat_table [TOTAL_MAT];

// Piece Square Table (PST) (+ 2 king endgame)
extern int pst [14][64];

//    - 
extern int att_score [15];

//      
extern const int att_weights [12];

extern int hi_king_att [256][256]; //      ( Rebel)
extern int hi_weak_att [256][256]; //     
extern int lo_king_att [256][256];
extern int lo_weak_att [256][256];

// 
extern int n_trop [64][64];
extern int b_trop [64][64];
extern int r_trop [64][64];
extern int q_trop [64][64];

//   
extern mate_t mate_pattern [65536];

// Reduction lookup tables (initialized at startup) and their getter functions
extern int reductions [2][64][64]; // [pv][depth][moveNumber]

//   
extern int un_castling [64];

extern int history [64][64];

extern int king_cast [2];
extern int queen_cast [2];

extern pv_t pv [PV_SIZE];

extern magic_t bishop_m [64];
extern magic_t rook_m  [64];

extern U64 bishop_att [64][512];
extern U64 rook_att  [64][4096];

extern U64 knight_att [64];
extern U64 king_att [64];
extern U64 link_att [64][64]; //    

extern U64 wpawn_att [64];
extern U64 bpawn_att [64];

extern int on_line [64][64];
extern int on_diag [64][64];

extern U64 n_mobility [64];
extern U64 b_mobility [64];
extern U64 r_mobility [64];

extern tune_t * points;
extern tune_t * dirs;
extern int val_point [USC_POINTS];
extern char str [2048];   //    





//////   ///////////////////////////////////////////////////////////////////////

// +------------------------+
// |  HASH.CPP              |
// +------------------------+
extern void    hash_init              ();
extern void    hash_refresh           (hash * _H);
extern void    reps_refresh           ();
extern void    pent_refresh           ();
extern hash_t* hash_get               (hash * _H, board * B);
extern pentry* pent_get               (board * B);
extern U64     hash_calc              (board * B);
extern void    hash_update            (int mv, int score, int type, int depth, hash * _H, board * B);
extern void    pent_update            (int score, eval_t * J, board * B);
extern void    hash_retrieve          (int &alpha, int &beta, int &mv, int depth, hash * _H, board * B);

// +------------------------+
// |  UCI.CPP               |
// +------------------------+
extern void    uci_send               (const char format[], ...);
extern void    uci_wait               ();
extern int     uci_parsing            (board * B);
extern void    uci_go_command         (int wtime, int btime, int absolute = 0, int fixdepth = 0, int infinite = 0);
extern void    uci_help               ();
extern void    uci_info               (board * B);
extern void    uci_pv_out             ();
extern void    uci_stop_searching     ();

// +------------------------+
// |  UTIL.CPP              |
// +------------------------+
extern void    init_mate_tables       ();
extern void    tropism_calc           (int king, int piece);

extern void    print_score            (int score);
extern void    shell_sort             (int * a, int * b, int num);
extern int     is_on_line             (int a, int b);

extern void    moves_to_pos           (char * fen, board * B);

extern char*   read_command           (char * str, const char * wrd);
extern int     read_number            (char * str);
extern char*   move_to_str            (int mv);
extern int     str_to_move            (char * s);

extern U64     rand64                 ();
extern U64     rand64_fewbits         ();

extern void    init_pst               ();

extern void    print_pv               (int len);
extern void    _sprint_pv             (char* pv_line, board * B);
extern void    sprint_pv              (int len, char* pv_line);

extern pv_t *  is_it_pv               (U64 key, int ply);
extern int     update_kihi            (int mv, int depth, board * B);
extern void    pv_write               (board * B, int move = 0);

extern U64     f_read64               (FILE* file);
extern U64     f_read_compact         (FILE* file);
extern int     f_write_compact        (FILE* file, U64 bb);

extern void    print_this_att         (U64 at);
extern int     print_move             (int mv);

extern bool    input_available        ();

// +------------------------+
// |  BOARD.CPP             |
// +------------------------+
extern void    board_init             ();

extern void    fen_to_pos             (const char * fen, board * B);
extern char*   pos_to_fen             (board * B);

extern U64     attacked_by            (U64 & occ, int sq);
extern U64     attacked_by_col        (U64 & occ, int sq, int col);
extern U64     attack_table           (U64 & occ, int sq, int col);
extern int     is_attacked            (U64 & occ, int sq, int col);

extern int     print_board            (board * B);

extern void    build_mask             (board * B);
extern void    check_detection        (board * B, int fast = 0);

extern int     mate_at_a_glance       (board * B);
extern int     threefold_repetition   (board * B);

// +------------------------+
// |  MOVES.CPP             |
// +------------------------+

extern int     set_att                (int col);
extern int     generate               (board * B);
extern int     generate_attacks       (board * B);
extern int     generate_quiet         (board * B);
extern int     generate_checks        (board * B);
extern int     generate_legal         (board * B);
extern int     legality_test          (int mv, board * B);
extern int     get_next_move          (board * B);
extern int     get_next_move_qs       (board * B);
extern int     correct_move           (int mv, board * B);
extern void    order_attacks          (board * B);
extern void    order_moves            (board * B);
extern void    order_moves_all        (board * B);
extern int     do_move                (int mv, int prom, board * B);
extern int     find_masked_move       (int mv, int mask, board * B);

extern int     remove_att             (int sq, board * B);
extern int     place_att              (int pi, int sq, board * B);

extern int     test_occ_pie_squ       (board * B);
extern int     make_move              (int mv, board * B, int self = 0); //    
extern void    unmake_move            (int mv, board * B);

extern void    make_null              (board * B);
extern void    unmake_null            ();

// +------------------------+
// |  EVAL.CPP              |
// +------------------------+
extern void    eval_init              ();
extern void    eval_reinit            ();
extern void    init_material_table    ();

extern int     calc_w_king_stage      (board * B);
extern int     calc_b_king_stage      (board * B);

extern int     eval_w_king            (board * B);
extern int     eval_b_king            (board * B);

extern int     eval_w_pawns           ();
extern int     eval_b_pawns           ();
extern int     eval_w_passers         ();
extern int     eval_b_passers         ();

extern int     eval_w_knights         ();
extern int     eval_b_knights         ();
extern int     eval_w_bishops         ();
extern int     eval_b_bishops         ();

extern int     eval_w_rooks           ();
extern int     eval_b_rooks           ();
extern int     eval_w_queens          ();
extern int     eval_b_queens          ();

extern int     eval_w_tactics         ();
extern int     eval_b_tactics         ();

extern void    calc_rook_sq           (int col);

extern void    eval_precalc           ();
extern void    soft_control           (int id, const U64 & att);
extern int     evaluate               (board * B, int alpha = -INF, int beta = INF);

extern void    debug_eval             ();

// +------------------------+
// |  SEARCH.CPP            |
// +------------------------+
extern void    search_init            ();
extern void    search_preinit         ();
extern int     iid                    (int m_depth, board * B);
extern int     search_pvs             (int alpha, int beta, int depth, bool in_pv, board * B);
extern int     q_search               (int alpha, int beta,            bool in_pv, board * B);
extern int     see                    (int mv);
extern S64     perft_root             (int depth, board * B);
extern S64     perft_ab               (int depth, board * B);

extern void    init_arrays            ();
extern void    init_uci_info          ();

// +------------------------+
// |  TUNING.CPP            |
// +------------------------+
extern void    init_constants         ();
extern void    save_constants         (tune_t * TT);
extern void    load_constants         ();

extern void    init_points            ();
extern int     tuning                 (char * name, board * B);
extern int     tuning_tg              (board * B);

extern void    to_lower               (char * str, int len = 0);
extern int     parse_tag              ();
extern int     parse_comment          ();
extern int     parse_moves            (char pre_char, board * B);

extern void    learning               (board * B);
extern int     err_func               (tune_t * T_OLD, tune_t * TT, board * B);
extern int     move_err_func          (int move, board * B);

extern void    point_clear            (tune_t * TT);
extern void    point_copy             (tune_t * TT, tune_t * TT_ORIG);
extern void    point_direct           (tune_t * TT, tune_t * DI, tune_t * TT_ORIG);
extern void    point_add_wgrad        (tune_t * TT, tune_t * DI, double weight);
extern int     point_norm             (tune_t * TT);
extern void    point_on_line          (double weight);

extern void    save_av_simplex        ();

extern void    shell_sort_tune        (double * a, tune_t * b, int num);


// +------------------------+
// |  MAGICS.CPP            |
// +------------------------+
extern void    init_attacks           ();
extern void    init_magics            ();
extern void    init_rook_magics       (int c);
extern void    init_bishop_magics     (int c);

extern U64     rmask                  (int sq);
extern U64     bmask                  (int sq);
extern U64     qmask                  (int sq);

extern U64     ratt                   (int sq, U64 block);
extern U64     batt                   (int sq, U64 block);

extern U64     index_to_64            (int index, int bits, U64 m);
