#include "main.h"

void search_preinit ()
{
    S->irreversible_ply = 0;
    S->fifty = 0;
	S->count_moves = 0;

	S->ready_to_search = 1; //    (  )
    S->deferred_cmd_go = 0; //   go
};

void search_init ()
{
    int i;

    S->g_hash_age = 0;
    S->B_start = &_B[0];

    S->state = ST_ANALYSING;
    S->infinite = 0;
    S->score = exc_draw (0);
    S->score_last = 0;
    S->best_move = 0;
    S->draw = 0;

    S->curr_time = 0;

    S->think_time = S->elaps_time = 0;

    S->pv_len = 0;
    S->pv_start = 0;

    S->nodes_count = E;
    S->nodes = E;

    S->ply = 0;
    S->search_depth = 0;

    for (i = 0; i < 8; i ++) S->mv_out [i] = 0;

    for (i = 0; i < PV_SIZE; i ++)
	{
	    pv [i].key = E;
	    pv [i].key_new = E;
        pv [i].mv = 0;
	};
};

// Internal Iterative Deepening + Search In Root
int iid (int m_depth, board * B)
{
    int i, k, sz;
    int val;

    int last_val = - INF;   //
    int last_move = 0;     //  time management
    int cnt_ch_moves = 0; //

    int singular = 0;    // 
    int best_val [2];

    int values [256];    //   ( mp)
	int alpha, beta, depth;
	int best_i [ROOT_ORDERING];

	for (i = 0; i < 256; i ++) values [i] = 0;

    S->state = ST_ANALYSING;
    S->best_move = 0; //    ( IID)

    //      
    init_arrays ();
    init_uci_info ();
    hash_calc (B);

	S->g_hash_age ++; //    aging'
	S->ply = 0;      //         

    //        
	sz = generate_legal (B);

    //    /   
	if (sz == 1)
	{
	    S->best_move = B->moves [0];
        S->score = exc_draw (0);

        uci_stop_searching ();

	    return 0;
	}
	else if (!sz) //   
	{
		if (B->check_is)
        {
            S->score = - INF;
            return 0;
        }
		else
        {
            S->draw = 1;
            S->score = DRAW;
            return 0;
        };
	};

    //    
    for (depth = 2; depth <= m_depth; depth ++)
    {
        #ifdef UCI_OUT
        if (depth >= DEPTH_TO_MOVE) uci_send ("info depth %i\n", depth);
        #endif

        S->search_depth = depth;
        S->nodes_count ++;

        alpha = - INF;
        beta = INF;

        for (i = 0; i < 256; i ++) values [i] = 0;
        for (i = 0; i < ROOT_ORDERING; i ++) best_i[i] = - 1;
        best_val[0] = best_val[1] = - INF;

        for (i = 0; i < sz; i ++)
        {
            B->cmove = B->moves[i];
            make_move (B->cmove, B);

            S->nodes_count ++;

            // info currmove c2c3 currmovenumber 17
            if (S->search_depth >= DEPTH_TO_CURRMV)
            {
                #ifdef UCI_OUT
                uci_send ("info currmove %s ", move_to_str (B->cmove));
                uci_send ("currmovenumber %i\n", i + 1);
                #endif
            };

            val = - search_pvs (- beta, - alpha, depth - 1, (i == 0), B + 1);

            unmake_move (B->cmove, B);

            if (S->state == ST_EXIT) return 0;

            if (val > alpha)
            {
                alpha = val;

                if (best_i [0] != i)
                {
                    for (k = ROOT_ORDERING - 1; k > 0; k --) best_i [k] = best_i [k - 1];
                    best_i [0] = i;

                    best_val [1] = best_val [0];
                    best_val [0] = val;
                };

                B->bmove = B->cmove; //    ( search_root)
                S->score = val;

                // -  pv
                S->pv_start = B->hash;
                pv_write (B);

                uci_pv_out ();

                if (S->state == ST_EXIT) return 0;
            };
        };

        val = alpha; // return alpha;

        //     
        for (k = 0; k < ROOT_ORDERING; k ++)
        {
            if (best_i[k] >= 0) values [best_i[k]] = ROOT_ORDERING - k;
        };

        if (S->state == ST_EXIT) return 0;

        S->best_move = B->bmove;

        if (S->best_move && is_mate (val)) break;

        // Singular Reply
        if (S->search_depth > 3
        &&  !S->infinite)
        {
            if (/*(best_val[1] == -INF && is_capture (S->best_move))
            ||  */(best_val[1] != -INF && absu (best_val[0], best_val[1]) > SINGULAR_VAL))
            {
                singular ++;
            }
            else singular = 0;

            if (singular >= 3)
            {
                S->best_move = B->moves [0];
                //S->score = 0;

                uci_stop_searching ();
                return 0;
            };
        };

        // Time management
        if (depth == 2)
        {
            last_val = val;
            last_move = S->best_move;
        };
        if (depth > 2)
        {
            if (last_move != S->best_move) //     
            {
                cnt_ch_moves ++;
                if (cnt_ch_moves > 4) S->is_critical = 1;
            };

            //if (absu (last_val, val) > 50) S->is_critical = 1;
            last_val = val;
            last_move = S->best_move;
        };

        if (depth >= DEPTH_TO_MOVE) uci_info (B); //    

        //     
        shell_sort ((int*)values, (int*)B->moves, sz);
    };

    uci_stop_searching ();
    return 0;
};


int search_pvs (int alpha, int beta, int depth, bool in_pv, board * B)
{
    int i, sz;
    int val, bval = - INF;
    int n_moves = 0;
    int o_alpha = alpha;
    int hmove = 0;
    //int refined = 0;
    int extend = 0;
    //int reduce = 0;
    int mate_threat = 0;
    int threat_move = 0;
    int dangerous = 0;
    int single_evasion = 0;
    //int have_7th_pawn = 0;
    //int non_pawn_mat = 0;
    int new_depth;
    bool search_pv = true; //     ,    ,  
    U64 x64;

    B->bmove = B->hmove = 0;
    B->stage = 0;

    S->nodes_count ++;

    //  ,    
    if (S->ply > MAX_DEPTH) return alpha;

    if (!((S->nodes_count + 1) & (NODES_OF_INPUT - 1)))
	{
	    if (!((S->nodes_count + 1) & (NODES_OF_SILENCE - 1))) uci_info (B);
	    uci_parsing (B);

	    if (S->state == ST_EXIT) return 0;
	};

    // Step 3. Mate distance pruning
    alpha = max(-INF + S->ply, alpha);
    beta = min(INF - (S->ply+1), beta);
    if (alpha >= beta) return alpha;

    // Step 4.     ,  
    if (!in_pv)
    {
        hash_retrieve (alpha, beta, hmove, depth, H, B);
        if (alpha == beta) return alpha;

        if (hmove)
        {
            bval = alpha;
            B->hmove = hmove; //   get_next_move ()
            B->bmove = hmove;
        };
    };

    check_detection (B);

    //refined = evaluate (B, alpha, beta);

	// Step 8. Null Move. Thanks to Igor Korshunov (Murka2 author) -----------------------------------
	if (!in_pv
    &&  !B->check_is
    &&  depth >= 1
    &&  beta < MATE
    &&  mat_for_null
    &&  !B->skip_null
    //&&  refined >= beta
    &&  USE_NULL_MOVE)
	{
	    // Null move dynamic reduction based on depth
        int R = 3 + (depth >= 5 ? depth / 8 : 0);

		make_null (B);

		(B+1)->skip_null = 1;

		if (depth > R) val = -search_pvs (-beta, -alpha, depth - R, 0, B + 1);
		else           val = -q_search   (-beta, -alpha, 0, B + 1);

        (B+1)->skip_null = 0;

		unmake_null ();

		if (val >= beta)
		{
		    //     (Stockfish)
		    if (val > MATE) val = beta;

		    if (depth < 6)
		    {
                hash_update (0, val, HASH_BETA, depth, H, B);
                return val;
		    };

            // Do verification search at high depths (Stockfish)
            B->skip_null = 1;
            int v = search_pvs (alpha, beta, depth - R, 0, B);
            B->skip_null = 0;

            if (v >= beta)
            {
                hash_update (0, val, HASH_BETA, depth, H, B);
                return val;
            };
		}
		else
		{
		    // The null move failed low, which means that we may be faced with
            // some kind of threat. If the previous move was reduced, check if
            // the move that refuted the null move was somehow connected to the
            // move which was reduced. If a connection is found, return a fail
            // low score (which will cause the reduced move to fail high in the
            // parent node, which will trigger a re-search with full depth).
            if (val == -INF + (S->ply + 2)) mate_threat = 1;

            threat_move = (B+1)->bmove;

            /*
            if (   depth < ThreatDepth
                && (ss-1)->reduction
                && connected_moves(pos, (ss-1)->currentMove, threatMove))
                return beta - 1;
            */
		};
	};

	// -------------------------------------------------------------------------------------------

    //check_detection (B); //  

    /////////////////////////////////////
    //
    //  Extensions
    //
    // 1.    

    //if ( B - 2 >= S->B_start
    //&&        B->check_is
    //&&  (B - 2)->check_is
    //&&  S->ply + depth < S->search_depth + CHECK_PLY_LIMIT) extend = 1;

    for (;;)
    {
        B->cmove = get_next_move(B);
        if (!(B->cmove)) break;
        if (!legality_test (B->cmove, B)) continue;

        make_move (B->cmove, B);

        n_moves ++;

        if (!threefold_repetition (B + 1) && S->fifty < 100)
        {
            single_evasion = B->check_is && (B->n_evasions == 1);
            dangerous = B->check_is || single_evasion || mate_threat;

            if (!extend)
            {
                //check_detection (B + 1);

                //  
                //if ((/*threat_move || */dangerous)
                //&&  S->ply + depth < S->search_depth + 6) extend = 1;

                //else

                //      -  
                //if ((B + 1)->check_is
                //&&  (B - 1)->check_is
                //&&  S->ply + depth < S->search_depth + CHECK_PLY_LIMIT) extend = 1;

                //else

                //      
                if (PIE (B->cmove) == WP
                &&  hori (FROM (B->cmove)) > 5
                &&  S->ply + depth < S->search_depth + PPUSH_PLY_LIMIT) extend = 1;

                else

                if (PIE (B->cmove) == BP
                &&  hori (FROM (B->cmove)) < 2
                &&  S->ply + depth < S->search_depth + PPUSH_PLY_LIMIT) extend = 1;
            };

            new_depth = depth - 1 + extend;


            if (in_pv      //  pv 
            &&  search_pv) //     ,  
            {
                if (new_depth > 0) val = - search_pvs (-beta, -alpha, new_depth, 1, B + 1);
                else               val = - q_search   (-beta, -alpha, 1, B + 1); // leaf node
            }
            else
            {
                // Step 14. Reduced depth search (non-pv)
                // If the move fails high will be re-searched at full depth.

                int do_full_search = 1;

                if (depth >= 3
                &&  !extend
                &&  !is_cap_or_prom (B->cmove)
                &&  !dangerous
                &&  !is_castling(B->cmove)
                &&  !move_is_killer(B->cmove, B))
                {
                    B->red = reduction (depth, n_moves, in_pv);
                    if (B->red)
                    {
                        int d = new_depth - B->red;

                        if (d > 0) val = -search_pvs (-alpha-1, -alpha, d, 0, B + 1);
                        else       val = -q_search   (-alpha-1, -alpha, 0, B + 1); // leaf node

                        do_full_search = (val > alpha);

                        //if (!do_full_search) uci_send (".");
                    };

                    // The move failed high, but if reduction is very big we could
                    // face a false positive, retry with a less aggressive reduction,
                    // if the move fails high again then go with full depth search.
                    if (do_full_search && B->red > 2)
                    {
                        B->red = 1;

                        val = -search_pvs (-(alpha+1), -alpha, new_depth - B->red, 0, B + 1);

                        do_full_search = (val > alpha);
                    };
                    B->red = 0; // Restore original reduction
                };

                // Step 15. Full depth search
                if (do_full_search)
                {
                    if (new_depth > 0) val = -search_pvs (-alpha-1, -alpha, new_depth, 0, B + 1);
                    else               val = -q_search   (-alpha-1, -alpha, 0, B + 1); // leaf node
                };

                /*
                if (new_depth > 0) val = -search_pvs (-alpha-1, -alpha, new_depth, 0, B + 1);
                else               val = -q_search   (-alpha-1, -alpha, 0, B + 1); // leaf node
                */
                if (val > alpha && val < beta)
                {
                    // re-search
                    if (new_depth > 0) val = -search_pvs (-beta, -alpha, new_depth, 0, B + 1);
                    else               val = -q_search   (-beta, -alpha, 0, B + 1); // leaf node
                };
            };
        }
        else val = DRAW;


        unmake_move (B->cmove, B);

        if (S->state == ST_EXIT) return 0;

        if (val > bval) // new best move
        {
            if (val > alpha) // alpha raising
            {
                alpha = val;
                search_pv = false;

                if (alpha >= beta) // cut-off
                {
                    //  ,   
                    hash_update (B->cmove, val, HASH_BETA, depth, H, B);
                    if (!is_capture(B->cmove) && !B->check_is) update_kihi (B->cmove, depth, B);

                    return alpha;
                };
            };

            bval = val;
            B->bmove = B->cmove;

            //  pv
            pv_write (B);
        };
    };

    //   

	if (!n_moves)
	{
		if (B->check_is) return - INF + S->ply;
		else            return 0;
	};

	if (bval > o_alpha) hash_update (B->bmove, alpha, HASH_EXACT, depth, H, B);
	else                hash_update (B->bmove, bval, HASH_ALPHA, depth, H, B);

    return alpha;
};

int q_search (int alpha, int beta, bool in_pv, board * B)
{
	int i, sz, val = 1;
	int hmove = 0;
	int depth = 0; // !   , ,    
	int bval;
	int check;
	int o_alpha = alpha;
	int n_moves = 0;
	bool search_pv = true; //     ,    ,  

    //return evaluate (B, alpha, beta);

    B->bmove = B->hmove = 0;
    B->stage = 0;

	S->nodes_count ++;

	if (!((S->nodes_count + 1) & (NODES_OF_INPUT - 1)))
    {
        if (!((S->nodes_count + 1) & (NODES_OF_SILENCE - 1))) uci_info (B);
        uci_parsing (B);

        if (S->state == ST_EXIT) return 0;
    };

    if (S->ply > MAX_DEPTH) return evaluate (B, alpha, beta);

    // Step 4.     ,  
    if (!in_pv)
    {
        hash_retrieve (alpha, beta, hmove, depth, H, B);
        if (alpha == beta) return alpha;

        if (hmove)
        {
            bval = alpha;
            B->hmove = hmove; //   get_next_move ()
            B->bmove = hmove;
        };
    };

    check_detection (B);

    if (!B->check_is)
	{
		bval = evaluate (B, alpha, beta);

		if (bval > alpha)
		{
			if (bval >= beta) return bval;
			alpha = bval;
		};
	}
	else
	{
		bval = - INF;
	};

    for (;;)
    {
        B->cmove = B->check_is ? get_next_move(B) : get_next_move_qs(B);
        if (!(B->cmove)) break;
        if (!legality_test (B->cmove, B)) continue;

		make_move (B->cmove, B);

		n_moves ++;

		if (!threefold_repetition (B + 1))
		{
		    if (in_pv && search_pv) val = - q_search (-beta, -alpha, 1, B + 1);
		    else
		    {
		        val = - q_search (-alpha - 1, -alpha, 0, B + 1);
		        if (val > alpha && val < beta) val = - q_search (-beta, -alpha, 0, B + 1);
		    };
		}
		else val = DRAW;

		unmake_move (B->cmove, B);

        if (S->state == ST_EXIT) return 0;

		if (val > bval) // new best move
		{
			if (val > alpha) // alpha raising
			{
			    search_pv = false;
			    alpha = val;

				//if (val >= beta)
				if (alpha >= beta) // beta cut-off
				{
				    //  
                    hash_update (B->cmove, val, HASH_BETA, depth, H, B);

				    return alpha;//val;
				};
			};

			bval = val;

			// -  pv
            pv_write (B);
		};
	};

    //      (  ) -   ,  
	//if (!n_moves && B->check_is) return - INF + S->ply;

	//   
	if (bval == - INF) bval += S->ply;

	if (bval > o_alpha) hash_update (B->bmove, alpha, HASH_EXACT, depth, H, B);
	else                hash_update (B->bmove, bval, HASH_ALPHA, depth, H, B);

	return alpha;//bval;
};

// Thanks to CPW
int see (int mv)
{
    int gain[32], d = 0;
    int _from, _to;
    int piece_from;
    int piece_to;
    int col;
    U64 rq, bq, pawns;
    U64 bb_from;
    U64 bb, t;
    U64 occ;

    _to        = (mv & 0x0000FF00) >> 8;
    _from      = (mv & 0x00FF0000) >> 16;
    piece_from = (mv & 0xFF000000) >> 24;
    piece_to   = P->squares [_to];

    bb_from    = I << _from;

    occ = P->occupied [0] | P->occupied [1];
    rq = bq = P->piece [BQ] | P->piece [WQ];
    rq     |= P->piece [BR] | P->piece [WR];
    bq     |= P->piece [BB] | P->piece [WB];
    pawns   = P->piece [BP] | P->piece [WP];

    bb = attacked_by (occ, _to); //   ,   

    gain [d] = material [piece_to];

    while (1)
    {
        d ++; //     
        gain [d] = material [piece_from] - gain [d - 1]; // ,  

        if (max (- gain [d - 1], gain [d]) < 0) break; //     

        bb  ^= bb_from; //     
        occ ^= bb_from; //      


        //  -      
        if      ( bb_from & pawns ) bb |= b_att (occ, _to) & occ & bq;
        else if ( bb_from & rq    ) bb |= q_att (occ, _to) & occ & rq; //   
        else if ( bb_from & bq    ) bb |= q_att (occ, _to) & occ & bq; //  

        col = d & 1;

        //    
        if      (t = bb & P->piece[BP + col]) bb_from = LSB(t); //    ?
        else if (t = bb & P->piece[BN + col]) bb_from = LSB(t); //   squares[]
        else if (t = bb & P->piece[BB + col]) bb_from = LSB(t);
        else if (t = bb & P->piece[BR + col]) bb_from = LSB(t);
        else if (t = bb & P->piece[BQ + col]) bb_from = LSB(t);
        else if (t = bb & P->piece[BK + col]) bb_from = LSB(t);
        else break;

        piece_from = P->squares [bitscan (bb_from)];
    };

    while (--d) gain [d - 1] = - max (- gain [d - 1], gain [d]);

    return gain [0];
};

S64 perft_root (int depth, board * B)
{
	int i, sz;
	int move;
	S64 perft = E;
	S64 all_nodes = E;
	clock_t curr_time;
	double perft_time;

	B->stage = 0;

    print_board (B);

    uci_send ("\n");

	if (!depth) return 1;

	check_detection (B);

    curr_time = clock ();

    for (;;)
	{
	    move = get_next_move(B);
	    if (!move) break;
		if (!legality_test (move, B)) continue;

		make_move (move, B);

		print_move (move);

		perft = perft_ab (depth - 1, B + 1);

		unmake_move (move, B);

		uci_send (" - %lli\n", perft);
        all_nodes += perft;
	};

	uci_send ("\n Totally %lli nodes\n", all_nodes);

    curr_time = clock () - curr_time;


    if (!curr_time) uci_send (" Generate time: less than 1/%i of second\n", CLOCKS_PER_SEC);
    else
    {
        perft_time = (double)curr_time / CLOCKS_PER_SEC;
        uci_send (" Generate time: %.3lf second(s)\n", (double)perft_time);
        uci_send (" %4.3lf knps\n\n", (all_nodes / perft_time) / 1000);
    };

	return perft;
};

S64 perft_ab (int depth, board * B)
{
    if (!depth) return 1;

	int i, sz, move;
	int cnt = 0;
	S64 perft = E;

	B->stage = 0;

	check_detection (B);

    for (;;)
	{
	    move = get_next_move(B);
	    if (!move) break;
		if (!legality_test (move, B)) continue;

	    //for (i = 0; i < depth; i ++) uci_send (" -");
	    //uci_send (" ");
        //print_move (move);
        //uci_send ("\n");

        if (depth == 1)
        {
            cnt ++;
            perft ++;
            continue;
        };

        make_move (move, B);

        cnt ++;

        perft += perft_ab (depth - 1, B + 1);

        unmake_move (move, B);
    };

    //   
    if (!cnt) return 0;

	return perft;
};

inline void init_arrays ()
{
    int i, j;

    for (i = 0; i < 64; i ++)
    {
        for (j = 0; j < 64; j ++)
        {
            history [i][j] = 0;
        };
    };
};

inline void init_uci_info ()
{
    clock_t clk;

    clk = clock ();

    S->think_time = clk;
    S->curr_time = clk;
};
