#include "main.h"

//    
#define TG_END   0             //  
#define TG_QUIT  0xFFFFFFFF    //   

//   
#define M_PIECE  0x7F000000
#define M_FR_HI  0x00380000
#define M_FR_LO  0x00070000
#define M_TO_HI  0x00003800
#define M_TO_LO  0x00000700
#define M_PROM   0x00000003
#define M_FR     (M_FR_LO | M_FR_HI)
#define M_TO     (M_TO_LO | M_TO_HI)

//      ,   ,     var (int, double)
#define write_var(a) fprintf (S->f_var, "%i ", a);
#define read_var(a)  fscanf (S->f_var, "%i", &a);


void init_constants () //  
{
    int i;

    //  
    TEMPO            = EINT(15); // 

    //  
    HOLES            = EINT(8);  //    
    BACKWARD         = EINT(17); //    
    BLOCKED_PAWN     = EINT(25); //      
    ISOLATED         = EINT(13); //    

    //  
    N_OUTPOST        = EINT(30); //    
    COLOR_WEAK       = EINT(16); //    
    FIANCHETTO       = EINT(10); //   
    BISHOP_PAIR      = EINT(15); //   2 
    RET_BISHOP       = EINT(10); //    
    BAD_BISHOP       = EINT(20); //    
    TRAPPED_BISHOP   = EINT(150); //    
    OPP_BISHOPS      = EINT(90); //   
    TRAPPED_ROOK     = EINT(170); //    
    BAD_ROOK         = EINT(40); //     (    )
    TARRASCH_RULE    = EINT(30); //      
    LINKED_ROOKS     = EINT(4);  //     -   (1/6  )
    ROOK_ON_OPENED   = EINT(35); //     
    ROOK_ON_SEMI     = EINT(22); //    
    ROOK_ON_7TH      = EINT(20); //    7 

    //  
    CONTACT_CHECK    = EINT(50); //  
    FORK             = EINT(20); // 

    //  
    CLOSED_KING      = EINT(15); //     (   )

    //  
    DISC_ATT         = EINT(15); // RNq
    DISC_CHECK       = EINT(17); // RNk
    PIN_PART         = EINT(8);  // Rrq
    PIN              = EINT(12); // Rnq
    PIN_ABS          = EINT(15); // Rnk
    SKEWER           = EINT(7);  // R(q|k)n -   
    XRAY             = EINT(5);  // RrN     - possible recapture,   

    //  
    CONTROL_P        = EINT(14)/16;
    CONTROL_N        = EINT(11)/16;
    CONTROL_B        = EINT( 9)/16;
    CONTROL_R        = EINT( 5)/16;
    CONTROL_Q        = EINT( 3)/16;
    CONTROL_K        = EINT( 2)/16;

    //   
    PST_MUL  (0)  = EINT(1);
    PST_MUL  (1)  = EINT(1);
    PST_MUL  (2)  = EINT(1);
    PST_MUL  (3)  = EINT(1);
    PST_MUL  (4)  = EINT(1);
    PST_MUL  (5)  = EINT(3);

    KING_DEV (0) = EINT(-50);
    KING_DEV (1) = EINT(-39);
    KING_DEV (2) = EINT(-27);
    KING_DEV (3) = EINT(-20);
    KING_DEV (4) = EINT(-15);
    KING_DEV (5) = EINT(-3);
    KING_DEV (6) = EINT(0);

    //    , , 
    SYNERGY (0)  = EINT(0);
    SYNERGY (1)  = EINT(4);
    SYNERGY (2)  = EINT(8);
    SYNERGY (3)  = EINT(20);
    SYNERGY (4)  = EINT(45);
    SYNERGY (5)  = EINT(100);
    SYNERGY (6)  = EINT(150);
    SYNERGY (7)  = EINT(170);

    // 
    MCENTER (0)  = EINT(-9);
    MCENTER (1)  = EINT(3);
    MCENTER (2)  = EINT(8);
    MCENTER (3)  = EINT(14);
    MCENTER (4)  = EINT(15);

    // 
    DOUBLED (0)   = EINT(15);
    DOUBLED (1)   = EINT(5);
    DOUBLED (2)   = EINT(8); // : 5..15
    DOUBLED (3)   = EINT(11);
    DOUBLED (4)   = EINT(11);
    DOUBLED (5)   = EINT(8);
    DOUBLED (6)   = EINT(5);
    DOUBLED (7)   = EINT(15);

    //  
	SHIELD (0)  = EINT(5);
	SHIELD (1)  = EINT(9);
	SHIELD (2)  = EINT(8);

	SHIELD (3)  = EINT(5);
	SHIELD (4)  = EINT(4);
	SHIELD (5)  = EINT(3);

	SHIELD (6)  = EINT(3);
	SHIELD (7)  = EINT(2);
	SHIELD (8)  = EINT(1);

	SHIELD (9)  = EINT(1);
	SHIELD (10) = EINT(1);
	SHIELD (11) = EINT(0);

    //  
	STORM (0)  = EINT(25);
	STORM (1)  = EINT(30);
	STORM (2)  = EINT(25);

	STORM (3)  = EINT(15);
	STORM (4)  = EINT(15);
	STORM (5)  = EINT(15);

	STORM (6)  = EINT(10);
	STORM (7)  = EINT(10);
	STORM (8)  = EINT(10);

	STORM (9)  = EINT(3);
	STORM (10) = EINT(5);
	STORM (11) = EINT(3);

    //    ()
    IMMOBILITY (0) = EINT (20); // 
    IMMOBILITY (1) = EINT (30); // 
    IMMOBILITY (2) = EINT (15); // 

    // 
    PASSER (1) = EINT(0);
    PASSER (2) = EINT(1);
    PASSER (3) = EINT(6);
    PASSER (4) = EINT(16);
    PASSER (5) = EINT(36);
    PASSER (6) = EINT(76);

    FRE_PASSER (1) = EINT(0);
    FRE_PASSER (2) = EINT(5);
    FRE_PASSER (3) = EINT(15);
    FRE_PASSER (4) = EINT(30);
    FRE_PASSER (5) = EINT(60);
    FRE_PASSER (6) = EINT(120);

    DEF_PASSER (1) = EINT(0);
    DEF_PASSER (2) = EINT(1);
    DEF_PASSER (3) = EINT(10);
    DEF_PASSER (4) = EINT(22);
    DEF_PASSER (5) = EINT(44);
    DEF_PASSER (6) = EINT(86);

    //          (   )
    NTROP (1) = EINT(12);
    NTROP (2) = EINT(6);
    NTROP (3) = EINT(4);
    NTROP (4) = EINT(2);
    NTROP (5) = EINT(1);

    BTROP (1) = EINT(10);
    BTROP (2) = EINT(4);

    RTROP (1) = EINT(8);
    RTROP (2) = EINT(7);

    QTROP (1) = EINT(7);
    QTROP (2) = EINT(2);
};

void save_constants (tune_t * TT)
{
    int i;

    S->f_var = fopen (F_CONFIG, "wb");

    for (i = 0; i < CONSTS_N; i ++) write_var (TT->v[i]);

    fclose (S->f_var);
};

void load_constants ()
{
    int i;

    S->f_var = fopen (F_CONFIG, "rb");

    if (S->f_var == NULL)
    {
        //fclose (S->f_var);
        save_constants (T);
        return;
    };

    for (i = 0; i < CONSTS_N; i ++) read_var (T->v[i]);

    fclose (S->f_var);
};

void init_points ()
{
    int i, j;

    for (i = 0; i < USC_POINTS; i ++)
    {
        for (j = 0; j < CONSTS_N; j ++)
        {
            points[i].v[j] = 0;
        };
    };
};

void learning (board * B)
{
    int i, j, k;
    int f[USC_POINTS];
    int fn;
    int a, b, c, d; //   
    int va, vb, vc, vd;

    uci_send ("\n");

    //    
    //   -   T0

    point_copy (&points[0], T0); //     

    while (1)
    {
        // 1.   ""    
        for (i = 1; i < USC_POINTS; i ++) point_direct (&points[i], &dirs[i], &points[0]);

        // 2.       
        uci_send ("1. New random directions\n");
        for (i = 0; i < USC_POINTS; i ++) f[i] = err_func (T0, &points[i], B);

        // 3.    ( )
        point_clear (&dirs[0]);
        for (i = 1; i < USC_POINTS; i ++) point_add_wgrad (&dirs[0], &dirs[i], f[i] - f[0]);

        // 4.   -
        fn = point_norm (&dirs[0]) + 1;

        uci_send ("\n2. Searching best point on the direction\n");

        // 5.       () a < b < c

        a = 0;
        d = USC_GAMMA;
        va = f[0]; //     
        point_on_line (d / (double (fn)));
        vd = err_func (T0, &points[1], B);
        for (i = 0; i < USC_DICHOT; i ++)
        {
            b = (a + d) / 2 - USC_DELTA;
            c = (a + d) / 2 + USC_DELTA;

            point_on_line (b / (double (fn)));
            vb = err_func (T0, &points[1], B);

            point_on_line (c / (double (fn)));
            vc = err_func (T0, &points[1], B);

            if (vb < vc) {a = b; va = vb;}
            else         {d = c; vd = vc;};
        };


        //  points[1]    
        point_copy (&points[0], &points[1]);

        //point_add_wgrad (&points[0], &dirs[0], USC_GAMMA / (double (fn)));

        // 6.   
        uci_send ("\n3. Saving result\n\n\n");
        save_constants (&points[0]);
    };
};

//  
inline void point_clear (tune_t * TT)
{
    for (int i = 0; i < CONSTS_N; i ++) TT->v[i] = 0;
};

//   ()
inline void point_copy (tune_t * TT, tune_t * TT_ORIG)
{
    for (int i = 0; i < CONSTS_N; i ++) TT->v[i] = TT_ORIG->v[i];
};

//           
void point_direct (tune_t * TT, tune_t * DI, tune_t * TT_ORIG)
{
    int i;
    double r[CONSTS_N];
    double sum = 0;

    for (i = 0; i < CONSTS_N; i ++)
    {
        r[i] = rand() / (double (RAND_MAX)) - .5;
        sum += r[i] * r[i];
    };

    sum = sqrt (sum);

    for (i = 0; i < CONSTS_N; i ++)
    {
        DI->v[i] = USC_RADIUS * (r[i] / sum);
        TT->v[i] = TT_ORIG->v[i] + DI->v[i];

        if (TT->v[i] < -EINT(WIN)) TT->v[i] = -EINT(WIN);
        if (TT->v[i] >  EINT(WIN)) TT->v[i] =  EINT(WIN);
    };
};

//   
inline void point_add_wgrad (tune_t * TT, tune_t * DI, double weight)
{
    for (int i = 0; i < CONSTS_N; i ++)
    {
        TT->v[i] += weight * DI->v[i];

        if (TT->v[i] < -EINT(WIN)) TT->v[i] = -EINT(WIN);
        if (TT->v[i] >  EINT(WIN)) TT->v[i] =  EINT(WIN);
    };
};

//    ()
int point_norm (tune_t * TT)
{
    long long sum = 0;
    for (int i = 0; i < CONSTS_N; i ++) sum += TT->v[i] * TT->v[i];

    return (int) sqrt(sum);
};

//       
inline void point_on_line (double weight)
{
    point_copy (&points[1], &points[0]);
    point_add_wgrad (&points[1], &dirs[0], weight);
};

//  .      
int err_func (tune_t * T_OLD, tune_t * TT, board * B)
{
    int i, result;
    int w_time, b_time;
    int w_tt = 1; //     
    int val[3];
    clock_t old, curr;

    val[0] = val[1] = val[2] = 0;

    S->tuning = 1;

    S->games = 0;
    S->tune_score = 0;

    for (i = 0; i < USC_TESTS; i ++, w_tt ^= 1)
    {
        //  
        search_init ();
        fen_to_pos (START_FEN, B);
        w_time = b_time = USC_TIME * CLOCKS_PER_SEC;
        result = 3; // 0 - , 1 - , 2 - , 3 - 

        //print_board (B);

        old = curr = clock ();

        while (1) //      !
        {
            //   ///////////////////////////////////////////////////////
            T = (P->wtm == w_tt) ? TT : T_OLD;
            eval_reinit ();
            uci_go_command (w_time, b_time, 0, USC_DEPTH);
            make_move(S->best_move, B, 1);

            curr = clock ();
            w_time -= curr - old;
            old = curr;
            if (w_time <= 0) //    ( )
            {
                result = 0;
                break;
            };

            //print_board (B);

            if (S->score == DRAW) result = 2;
            else if (S->score > USC_WIN_SCORE) result = 1;
            else if (S->score < - USC_WIN_SCORE) result = 0;

            //   //////////////////////////////////////////////////////
            T = (P->wtm == w_tt) ? TT : T_OLD;
            eval_reinit ();
            uci_go_command (w_time, b_time, 0, USC_DEPTH);
            make_move(S->best_move, B, 1);

            curr = clock ();
            b_time -= curr - old;
            old = curr;
            if (b_time <= 0) //    ( )
            {
                result = 1;
                break;
            };

            //print_board (B);

            if (result == 1) //   ( )  
            {
                if (S->score >= - USC_WIN_SCORE) result = 3;
                else break;
            }
            else if (result == 0) //      
            {
                if (S->score <= USC_WIN_SCORE) result = 3;
                else break;
            }
            else if (result == 2) //    
            {
                if (S->score != DRAW) result = 3;
                else break;
            };
        };

        if (w_tt == result)            {val[0] ++; uci_send ("+");}
        else if ((w_tt ^ 1) == result) {val[1] ++; uci_send ("-");}
        else                           {val[2] ++; uci_send ("=");};
    };

    uci_send (" |  + %-4i = %-4i - %-4i", val[0], val[2], val[1]);

    // ,  val[2] < USC_TESTS,    
    // :    -  
    S->tune_score = 2 * USC_TESTS * (val[0] - val[1] + USC_TESTS) - val[2];

    uci_send (" score: %5i\n", S->tune_score);

    return S->tune_score;
};

//   
int _err_func (tune_t * T_OLD, tune_t * TT, board * B)
{
    int i;
    long long r = 0;

    for (i = 0; i < CONSTS_N; i ++)
    {
        r += (TT->v[i] - T_OLD->v[i] - 10) * (TT->v[i] - T_OLD->v[i] - 10);
    };

    S->tune_score = 100000 - (long long)sqrt(r);
    uci_send (" score: %5i\n", S->tune_score);

    return S->tune_score;
};

void shell_sort_tune (double * a, tune_t * b, int num)
{
    register int i, j, k, inc;
    double x;
    tune_t y;

    int arr[] = {9, 5, 3, 2, 1};

    for (k = 0; k < 5; k ++)
    {
        inc = arr[k];
        for (i = inc; i < num; ++ i)
        {
            x = a[i];
            y = b[i];

            for (j = i - inc; (x < a[j]) && (j >= 0); j -= inc)
            {
                a[j + inc] = a[j];
                b[j + inc] = b[j];
            };

            a[j + inc] = x;
            b[j + inc] = y;
        };
    };
};








