#include "main.h"



void uci_send (const char format[], ...) // Fruit
{
    va_list  arg_list;
    char     str [4096];

    va_start (arg_list, format);
    vsprintf (str, format, arg_list);
    va_end   (arg_list);

    fprintf  (stdout, "%s", str);

    #ifdef LOGGING

    if (strlen(str) > 0) fprintf (S->f_out, ">>%s\n", str);

    #endif

    fflush   (stdout);
};

void uci_wait ()
{
    char * ptr;
	char s[2048] = "";

	for (;;)
	{
		fgets (s, 2040, stdin);
		if (feof(stdin)) continue;

		ptr = strchr (s, '\n'); //    
        if (ptr != NULL) *ptr = '\0';

        #ifdef LOGGING
		if (strlen (s) > 0) fprintf (S->f_out, "[[%s", s);
		#endif

		if (read_command (s, "quit") != NULL)
		{
		    #ifdef LOGGING
			fclose (S->f_out);
			#endif
			exit (0);
		};
		if (read_command (s, "uci") != NULL)
		{
			uci_send ("id name %s v%s\n", NAME, VERSION);
			uci_send ("id author %s\n", AUTHOR);

			uci_send ("option name Hash type spin default 1 min 1 max 128\n");
			//uci_send ("option name EvalVars type string default default.var\n");

			uci_send ("uciok\n");
			break;
		};
		if (read_command (s, ".") != NULL)
		{
			//debug = 1;
			//break;
		};
	};
};

int uci_parsing (board * B)
{
    char *ptr, *d, *f;

	int j, k;
	//int score;
	int infinite; //  
	int fixdepth; //  
	int absolute; //   (  )
	int wtime, btime;
    U64 perft;

	while (!S->deferred_cmd_go && S->state != ST_EXIT)
	{
	    if (!input_available()) break;

        fgets (str, 2040, stdin);

        ptr = strchr (str, '\n'); //   
        if (ptr != NULL) *ptr = '\0';

        //if (feof (stdin)) break;

        #ifdef LOGGING

        if (strlen(str) > 0) fprintf (S->f_out, "[[%s\n", str);

        #endif

        //     
        if (strlen(str) > 0) switch (str[0])
        {
            case 'q': // quit

                if (read_command (str + 1, "uit") != NULL)
                {
                    #ifdef LOGGING
                    fclose (S->f_out);
                    #endif
                    S->state = ST_EXIT;
                    return 1;
                };

                break;

            case 'i': // isready

                if (read_command (str + 1, "sready") != NULL)
                {
                    uci_send ("readyok\n");
                    break;
                };

                break;

            case 'p': // position

                if ((d = read_command (str + 1, "osition ")) != NULL)
                {
                    if ((f = read_command (d, "fen ")) != NULL)
                    {
                        fen_to_pos (f, B);

                        if ((d = read_command (f, "moves ")) != NULL)
                        {
                            moves_to_pos (d, B);
                        };
                        break;
                    };

                    if ((f = read_command (d, "startpos")) != NULL)
                    {
                        fen_to_pos (START_FEN, B);

                        if ((d = read_command (f, " moves ")) != NULL)
                        {
                            moves_to_pos (d, B);
                        };
                        break;
                    };
                };

                break;

            case 'g': // go

                if ((d = read_command (str + 1, "o")) != NULL)
                {
                    wtime = btime = 100000;

                    infinite = 0;
                    if ((read_command (d, "infinite")) != NULL)
                    {
                        infinite = 1;
                    };

                    fixdepth = 0;
                    if ((f = read_command (d, "depth ")) != NULL)
                    {
                        fixdepth = read_number (f);
                    };

                    absolute = 0;
                    if ((f = read_command (d, "movetime ")) != NULL)
                    {
                        absolute = read_number (f);
                    };

                    //    wtime  btime !
                    if ((f = read_command (d, "wtime ")) != NULL)
                    {
                        wtime = read_number (f);

                        if ((f = read_command (d, "btime ")) != NULL)
                        {
                            btime = read_number (f);
                        };
                    };

                    uci_go_command (wtime, btime, absolute, fixdepth, infinite);

                    //hash_refresh ();

                    break;
                };

                break;

            case 's':

                if (str[1] == 'e') // setoption name
                {
                    if ((d = read_command (str + 2, "toption name ")) != NULL)
                    {
                        if ((f = read_command (d, "Hash ")) != NULL)
                        {
                            if ((d = read_command (f, "value ")) != NULL)
                            {
                                j = read_number (d);

                                if (j > 0 && j <= 512) //    512  
                                {
                                    delete H;
                                    H = new hash(j * (1 << 20));
                                }
                                else uci_send ("info string I can't allocate more than 512Mb memory for hash\n");
                            };
                        };

                        break;
                    };
                }
                else if (str[1] == 't') // stop
                {
                    if (read_command (str + 2, "op") != NULL)
                    {
                        S->state = ST_EXIT;
                        print_score (S->score);
                        uci_send ("pv %s\n", move_to_str (S->best_move));
                        uci_send ("bestmove %s \n", move_to_str (S->best_move));

                        break;
                    };
                };

                break;

            case 'h': // help

                if (read_command (str + 1, "elp") != NULL)
                {
                    uci_help ();
                    break;
                };

                break;

            case '.':

                if ((d = read_command (str + 1, "perft")) != NULL)
                {
                    j = 1;

                    if ((f = read_command (d, " ")) != NULL) if (k = read_number (f)) j = k;

                    perft_root (j, S->B_start);

                    break;
                };

                if (read_command (str + 1, "bench") != NULL)
                {
                    for (j = BENCH_LW_DEPTH; j <= BENCH_HG_DEPTH; j ++)
                    {
                        perft = perft_root (j, S->B_start);
                    };

                    break;
                };

                if (read_command (str + 1, "print") != NULL)
                {
                    print_board (B);

                    break;
                };

                if (read_command (str + 1, "score") != NULL)
                {
                    print_board (B);
                    uci_send (" Score: %i\n\n", evaluate (B, -INF, INF));

                    break;
                };

                if (read_command (str + 1, "fenout") != NULL)
                {
                    uci_send ("%s to move\n", P->wtm ? "White" : "Black");
                    uci_send ("fen: %s\n", pos_to_fen (B));

                    break;
                };

                break;

            default:

                uci_help ();
                break;
        };
	};

	return 0;
};

void uci_go_command (int wtime, int btime, int absolute, int fixdepth, int infinite)
{
    if (!S->ready_to_search)
    {
        S->deferred_cmd_go = 1;
        return;
    };

    S->ready_to_search = 0; //   -        
    //hash_refresh (H);
    //hash_refresh (Q);
    search_init ();

    S->infinite = infinite;
    S->state = ST_ANALYSING;
    S->sudden_death = 0;

    if (!absolute)
    {
        S->critl_time = S->elaps_time = P->wtm ? wtime : btime;

        if (S->elaps_time)
        {
            if (S->elaps_time < SUDDEN_DEATH_TIME * CLOCKS_PER_SEC) // SUDDEN DEATH
            {
                S->sudden_death = 1;
            }
            else
            {
                S->elaps_time = S->elaps_time / 50; //   (/50)
                S->critl_time = S->critl_time / 25; //  e (/15)
            };
        };
    }
    else S->critl_time = S->elaps_time = absolute;

    S->is_critical = 0;

    iid (fixdepth ? fixdepth : GO_MAX_DEPTH, S->B_start);

    S->ready_to_search = 1; //     -    
    S->deferred_cmd_go = 0; //    
};

inline void uci_help ()
{
    uci_send ("\n");
    uci_send (" setoption name [NAME] value [VALUE]             - set option state (UCI)\n");
    uci_send (" position (fen [FEN] (moves [MOVES])) (startpos) - set new position (UCI)\n");
    uci_send (" go (wtime [TIME]) (btime [TIME])                - start to analyse (UCI)\n");
    uci_send (" isready                                         - don't sleep!     (UCI)\n");
    uci_send (" quit                                            - exit             (UCI)\n\n");
    uci_send (" .perft ([DEPTH])                                - performance test \n");
    uci_send (" .bench                                          - benchmark based on perft\n");
    uci_send (" .print                                          - print out position\n");
    uci_send (" .score                                          - position score\n");
    uci_send (" .fenout                                         - print out fen\n\n");
};

void uci_info (board * B)
{
    long    nps; //    
    clock_t curr, diff;

    //if (S->tuning) return;
    if (S->state != ST_ANALYSING) return;

    curr = clock ();

    //      ,  -  ,   
    if (!S->infinite
    &&  S->best_move
    &&  S->search_depth >= DEPTH_TO_MOVE)
    {
        if (S->sudden_death)
        {
            uci_stop_searching ();
            return;
        }
        else if (curr - S->think_time > S->elaps_time)
        {
            if (S->is_critical)
            {
                if (curr - S->think_time > S->critl_time)
                {
                    uci_stop_searching ();
                    return;
                };
            }
            else
            {
                uci_stop_searching ();
                return;
            };
        };
    };

    //      
    diff    =  curr - S->curr_time;

    if (!diff) nps = 0;
    else       nps = (S->nodes_count * CLOCKS_PER_SEC) / diff;

    S->nodes += S->nodes_count;

    #ifdef UCI_OUT
    // info time 42219 nodes 8490000 nps 201094 cpuload 1000
    uci_send ("info time %i nodes %lli nps %li cpuload 1000\n", (int)(curr - S->think_time), S->nodes, nps);
    #endif

    S->curr_time = curr;
    S->nodes_count = E;
};

void uci_pv_out ()
{
    char pv_line [MAX_DEPTH * 10] = {0, };

    #ifdef UCI_OUT

    //   
    if (S->score_last != S->score
    &&  S->search_depth >= DEPTH_TO_MOVE)
    {
        S->score_last = S->score;

        print_score (S->score);

        #ifdef PV_OUTPUT
        sprint_pv (S->search_depth, pv_line);
        uci_send ("pv %s\n", pv_line);
        #else
        uci_send ("pv %s\n", move_to_str (S->best_move));
        #endif
    };

    #endif
};

void uci_stop_searching ()
{
    S->state = ST_EXIT;
    #ifdef UCI_OUT
    print_score (S->score);
    uci_send ("pv %s\n", move_to_str (S->best_move));
    uci_send ("bestmove %s \n", move_to_str (S->best_move));
    #endif
    S->nodes = E;
};
