#include "main.h"

void init_mate_tables ()
{
    int key;
    int i, j;
    int beyond;
    int disc_nq;
    int disc_rbq;
    int close;
    U64 blc, att, bbb;
    U64 bb, t, map;
    U64 safety;
    U64 x64;

    //       
    static const int to_board [8] = {0, 1, 2, 8, 10, 16, 17, 18};

    safety = king_att [B2];

    //   ,  
    for (i = 0; i < 65536; i ++)
    {
        blc = att = E;
        beyond = close = 0;
        disc_nq = disc_rbq = 0;

        for (key = i & 0xFF; key; key ^= t)
        {
            t = LSB (key);
            j = bitscan (t);

            att |= I << to_board [j];
        };

        for (key = i >> 8; key; key ^= t)
        {
            t = LSB (key);
            j = bitscan (t);

            blc |= I << to_board [j];
        };

        //      
        //     

        //   ,   ,
        //      
        bb = blc | att;


        ///////////////////////////////////////////////////////////////////
        //
        //  Step 1. Beyond mate -   
        //
        if (!(blc & _A2))
        {
            //    
            map = safety & ~_A2 & ~_C2;
            if ((bb & map) == map) beyond |= MATE_A2;
            else
            {
                //   
                if (pop_count (bb ^ map) <= 2)
                {
                    //  1

                };

            };
        };

        if (!(blc & _B3))
        {
            map = safety & ~_B3 & ~_B1;
            if ((bb & map) == map) beyond |= MATE_B3;
        };

        if (!(blc & _C2))
        {
            map = safety & ~_A2 & ~_C2;
            if ((bb & map) == map) beyond |= MATE_C2;
        };

        if (!(blc & _B1))
        {
            map = safety & ~_B3 & ~_B1;
            if ((bb & map) == map) beyond |= MATE_B1;
        };

        // 

        if (!(blc & _A1))
        {
            map = safety & ~_A1 & ~_C3;
            if ((bb & map) == map) beyond |= MATE_A1;
        };

        if (!(blc & _A3))
        {
            map = safety & ~_A3 & ~_C1;
            if ((bb & map) == map) beyond |= MATE_A3;
        };

        if (!(blc & _C1))
        {
            map = safety & ~_C1 & ~_A3;
            if ((bb & map) == map) beyond |= MATE_C1;
        };

        if (!(blc & _C3))
        {
            map = safety & ~_C3 & ~_A1;
            if ((bb & map) == map) beyond |= MATE_C3;
        };

    };
};

void tropism_calc (int king, int piece)
{
    int j, ply;
    U64 att, bb, t;
    U64 saved_att;
    U64 dont_looked;

    att = I << king; //    0
    dont_looked = U ^ att;

    for (ply = 1; ply < 20; ply ++)
    {
        bb = att; //   ply - 1
        att = E; //   ply

        //   
        while (bb)
        {
            t = LSB (bb);
            j = bitscan (t);

            switch (piece)
            {
                case KNIGHT_ID : att |= knight_att [j]; break;
                case BISHOP_ID : att |= b_att (E, j);   break;
                case ROOK_ID   : att |= r_att (E, j);   break;
                case QUEEN_ID  : att |= q_att (E, j);   break;
            };

            bb ^= t;
        };

        att &= dont_looked; //   ...
        dont_looked ^= att; //                   ... 

        saved_att = att;

        while (att)
        {
            t = LSB (att);
            j = bitscan (t);

            //    
            switch (piece)
            {
                case KNIGHT_ID : n_trop [king][j] = ((ply < 6) ? NTROP (ply) : 0);                         break;
                case BISHOP_ID : b_trop [king][j] = ((ply < 3) ? BTROP (ply) : 0) - (dist (king, j) >> 2); break;
                case ROOK_ID   : r_trop [king][j] = ((ply < 3) ? RTROP (ply) : 0) - (dist (king, j) >> 1); break;
                case QUEEN_ID  : q_trop [king][j] = ((ply < 3) ? QTROP (ply) : 0) - (dist (king, j) >> 2); break;
            };

            att ^= t;
        };

        att = saved_att;
    };

    //  
    switch (piece)
    {
        case KNIGHT_ID : n_trop [king][king] = 0; break;
        case BISHOP_ID : b_trop [king][king] = 0; break;
        case ROOK_ID   : r_trop [king][king] = 0; break;
        case QUEEN_ID  : q_trop [king][king] = 0; break;
    };
};

void moves_to_pos (char *fen, board * B)
{
    int mv;
    int prom = 0; //  - ?

    while (*fen)
    {
        while (*fen == ' ') fen ++;

        if (!(*fen)) return;

        mv = str_to_move (fen);
        if (!mv) return;

        if (mv & f_prom)
        {
            prom = 1;
        };

        if (!do_move (mv, prom, B))
        {
            uci_send ("info string Can't understand move line\n");

            print_board (B);

            uci_send ("%s\n", fen);

            return;
        };

        while (*fen != ' ' && *fen) fen ++;
        //fen += 4; //     4 !
    };
};

char* read_command (char *str, const char *wrd)
{
	for (;; str++)
	{
		if (!*str) return NULL;
		if (*str == *wrd) break;
	};

	for (;; str++, wrd++)
	{
	    if (!*wrd) return str; //  ,    - 
	    if (!*str || !*wrd) return NULL;
		if (*str != *wrd) return NULL;
	};

	return str;
};

int read_number (char *str)
{
	char ch;
	int num = 0;

	for (;; str ++)
	{
		ch = *str;
		if (ch < '0' || ch > '9') break;

		num *= 10;
		num += ch - '0';
	};

	return num;
};

char* move_to_str (int mv)
{
    char * buff = S->mv_out;
    int _flags, _to, _from;

    _flags = mv & 0xFF;
    mv >>= 8;

    _to    = mv & 0xFF;
    mv >>= 8;

    _from  = mv & 0xFF;
    mv >>= 8;

    sprintf (buff    , "%c%c", 'a' + (vert (_from)), '1' + (hori (_from)));
    sprintf (buff + 2, "%c%c", 'a' + (vert (_to))  , '1' + (hori (_to)));

    if (is_prom (_flags)) sprintf (buff + 4, "%c\0", "nbrq"[_flags & ~f_mask]);
    else                  sprintf (buff + 4, "\0");

    return buff;
}

int str_to_move (char * s)
{
    char a, b, c, d, e;
    int _from, _to, _flags;

    a = s[0] - 'a';
    b = s[1] - '1';
    c = s[2] - 'a';
    d = s[3] - '1';
    e = s[4];

    //uci_send ("prepared to promote: %c\n", e);

    _flags = 0;

    switch (e)
    {
        case 'n': _flags = f_prom + 0; break;
        case 'b': _flags = f_prom + 1; break;
        case 'r': _flags = f_prom + 2; break;
        case 'q': _flags = f_prom + 3; break;
        case ' ': _flags = 0; break;
        default : _flags = 0; break; //   - 
    };

    if (a < 0 || a > 7) return 0;
    if (b < 0 || b > 7) return 0;
    if (c < 0 || c > 7) return 0;
    if (d < 0 || d > 7) return 0;

    _from = compose (b, a);
    _to   = compose (d, c);

    //if ((b == d) && (absu (a, c) == 2)) _flags |= f_castling;

    return (_from << 16) | (_to << 8) | _flags;
};

void init_pst ()
{
    int i;
    int sq;

    //       ( ):
/*
                                A1, B1, C1, D1, E1, F1, G1, H1,
                                A2, B2, C2, D2, E2, F2, G2, H2,
                                A3, B3, C3, D3, E3, F3, G3, H3,
                                A4, B4, C4, D4, E4, F4, G4, H4,
                                A5, B5, C5, D5, E5, F5, G5, H5,
                                A6, B6, C6, D6, E6, F6, G6, H6,
                                A7, B7, C7, D7, E7, F7, G7, H7,
                                A8, B8, C8, D8, E8, F8, G8, H8};
*/
    const int pst_p      [] = {  0,  0,  0,  0,  0,  0,  0,  0,
                                50, 50, 50, 50, 50, 50, 50, 50,
                                10, 10, 20, 30, 30, 20, 10, 10,
                                 5,  5, 10, 25, 25, 10,  5,  5,
                                 0,  0,  0, 20, 20,  0,  0,  0,
                                 5, -5,-10,  0,  0,-10, -5,  5,
                                 5, 10, 10,-20,-20, 10, 10,  5,
                                 0,  0,  0,  0,  0,  0,  0,  0};

    const int pst_n      [] = {-30,-25,-20,-20,-20,-20,-25,-30,
                               -25,-20,  0,  0,  0,  0,-20,-25,
                               -20,  0, 10, 15, 15, 10,  0,-20,
                               -20,  5, 15, 20, 20, 15,  5,-20,
                               -20,  0, 15, 20, 20, 15,  0,-20,
                               -20,  5, 10, 15, 15, 10,  5,-20,
                               -25,-20,  0,  5,  5,  0,-20,-25,
                               -30,-25,-20,-20,-20,-20,-25,-30};

    const int pst_b      [] = {-20,-10,-10,-10,-10,-10,-10,-20,
                               -10,  0,  0,  0,  0,  0,  0,-10,
                               -10,  0,  5, 10, 10,  5,  0,-10,
                               -10,  5,  5, 10, 10,  5,  5,-10,
                               -10,  0, 10, 10, 10, 10,  0,-10,
                               -10, 10, 10, 10, 10, 10, 10,-10,
                               -10,  5,  0,  0,  0,  0,  5,-10,
                               -20,-10,-10,-10,-10,-10,-10,-20};

    const int pst_r      [] = {  0,  0,  0,  0,  0,  0,  0,  0,
                                -5,  0,  0,  0,  0,  0,  0, -5,
                                -5,  0,  0,  0,  0,  0,  0, -5,
                                -5,  0,  0,  0,  0,  0,  0, -5,
                                -5,  0,  0,  0,  0,  0,  0, -5,
                                -5,  0,  0,  0,  0,  0,  0, -5,
                                -5,  0,  0,  0,  0,  0,  0, -5,
                                 0,  0,  0,  5,  5,  0,  0,  0};

    const int pst_q      [] = {-20,-10,-10, -5, -5,-10,-10,-20,
                               -10,  0,  0,  0,  0,  0,  0,-10,
                               -10,  0,  5,  5,  5,  5,  0,-10,
                                -5,  0,  5,  5,  5,  5,  0, -5,
                                 0,  0,  5,  5,  5,  5,  0, -5,
                               -10,  5,  5,  5,  5,  5,  0,-10,
                               -10,  0,  5,  0,  0,  0,  0,-10,
                               -20,-10,-10, -5, -5,-10,-10,-20};

    // King in middle game
    const int pst_km     [] = { -3, -4, -4, -5, -5, -4, -4, -3,
                                -3, -4, -4, -5, -5, -4, -4, -3,
                                -3, -4, -4, -5, -5, -4, -4, -3,
                                -3, -4, -4, -5, -5, -4, -4, -3,
                                -2, -3, -3, -4, -4, -3, -3, -2,
                                -1, -2, -2, -2, -2, -2, -2, -1,
                                 2,  2,  0,  0,  0,  0,  2,  2,
                                 2,  3,  1,  0,  0,  1,  3,  2};

    // King in end game
    const int pst_ke     [] = { -5, -4, -3, -2, -2, -3, -4, -5,
                                -3, -2, -1,  0,  0, -1, -2, -3,
                                -3, -1,  2,  3,  3,  2, -1, -3,
                                -3, -1,  3,  4,  4,  3, -1, -3,
                                -3, -1,  3,  4,  4,  3, -1, -3,
                                -3, -1,  2,  3,  3,  2, -1, -3,
                                -3, -3,  0,  0,  0,  0, -3, -3,
                                -5, -3, -3, -3, -3, -3, -3, -5};

    //  (  chessprogramming,   )
    for (sq = 0; sq < 64; sq ++)
    {
        pst [WP][sq]     = PST_MUL(0) * pst_p  [sq^070];// * 4 / 5;
        pst [WN][sq]     = PST_MUL(1) * pst_n  [sq^070];// * 3 / 5;
        pst [WB][sq]     = PST_MUL(2) * pst_b  [sq^070];// * 4 / 5;
        pst [WR][sq]     = PST_MUL(3) * pst_r  [sq^070];// * 2 / 5;
        pst [WQ][sq]     = PST_MUL(4) * pst_q  [sq^070];// * 2 / 5;
        pst [WK][sq]     = PST_MUL(5) * pst_km [sq^070];
        pst [WK + 2][sq] = PST_MUL(5) * pst_ke [sq^070]; //    
    };

    //    
    for (sq = 0; sq < 64; sq ++)
    {
        for (i = 0; i < 14; i += 2)
        {
            pst [i][sq] = pst [i^1][sq^070];
        };
    };

    /*
    for (i = 7; i >= 0; i --)
    {
        for (int j = 0; j < 8; j ++)
        {
            sq = (i << 3) + j;
            printf ("%4i ", pst [BK + 2][sq]);
        };
        printf ("\n");
    };
    */
};

U64 rand64 ()
{
    U64 u1, u2, u3, u4;

	u1 = (U64)(rand()) & 0xFFFF;
	u2 = (U64)(rand()) & 0xFFFF;
	u3 = (U64)(rand()) & 0xFFFF;
	u4 = (U64)(rand()) & 0xFFFF;

	return u1 | (u2 << 16) | (u3 << 32) | (u4 << 48);
}

U64 rand64_fewbits()
{
	return rand64() & rand64() & rand64();
};

void print_pv (int len)
{
	int i;
	U64 key;
	int mv;

	for (key = S->pv_start, i = key & (PV_SIZE - 1); key != NULL;)
	{
	    i = key & (PV_SIZE - 1);
	    while (pv[i].mv && key != pv[i].key) i = (i < PV_SIZE - 1) ? (i + 1) : 0;

	    mv = pv [i].mv;
	    if (!mv) break;

		uci_send ("%s ", move_to_str (mv));

		key = pv[i].key_new;
		if (key == S->pv_start) break;
	};

	uci_send ("\n");
};

void _sprint_pv (char* pv_line, board * B)
{
	int   i, j;
	U64   key;
	char* move;
	char  ch;
	int   mv;
	U64   keys [128];
	int   movs [128];
	int   curr_key = 0;
	hash_t * h;

	if (S->search_depth == 8)
	{
	    S = S + 1 - 1;
	};

    for (key = S->pv_start, i = key & (H->size - 1); key != NULL;)
	{
	    //        
	    i = key & (H->size - 1);
	    h = &H->entry [i];

        //       
        mv = h->move;
        movs[curr_key] = mv;

	    if (!mv) break;

		move = move_to_str (mv);

        //    pv-
		while (ch = *move++) *pv_line++ = ch;
		*pv_line++ = ' ';

        //  -    
        make_move (mv, B, 1);
		key = B->hash;
		if (key == S->pv_start) break;

		keys [curr_key ++] = key;
		if (curr_key > 40) break;
	};

    //    
	for (i = curr_key - 1; i >= 0; i --) unmake_move (movs[i], B);

	*pv_line = 0;
};

void sprint_pv (int len, char* pv_line)
{
	int   i, j;
	U64   key;
	char* move;
	char  ch;
	int   mv;
	U64   keys [128];
	int   curr_key = 0;

    for (key = S->pv_start, i = key & (PV_SIZE - 1); key != NULL;)
	{
	    i = key & (PV_SIZE - 1);
	    while (pv[i].mv && key != pv[i].key) i = (i < PV_SIZE - 1) ? (i + 1) : 0;

        mv = pv [i].mv;
        if (!mv) break;

        if (curr_key)
        {
            for (j = 0; j < curr_key - 1; j ++)
            {
                if (key == keys[j])
                {
                    mv = 0;
                    break;
                };
            };
        };

	    if (!mv) break;

        //printf ("%i\n", mv);
		move = move_to_str (mv);

		while (ch = *move++) *pv_line++ = ch;

		key = pv[i].key_new;
		if (key == S->pv_start) break;

		*pv_line++ = ' ';

		keys [curr_key ++] = key;
		if (curr_key > len) break;
	};

	*pv_line = 0;
};

U64 f_read64 (FILE* file)
{
	int t;
	U64 bb, b;
	U32 p;

	t = fscanf (file, "%8x", &p);
	b = p;
	bb = b << 32;

	if (t == -1)
	{
		S->err = 1;
		return U;
	};

	t = fscanf (file, "%8x", &p);
	b = p;
	bb += b & 0xFFFFFFFF;

	if (t == -1)
	{
		S->err = 1;
		return U;
	};

	return bb;
};

U64 f_read_compact (FILE* file)
{
	int i, t;
	U64 bb;
	uchar ch;

	bb = E;

    for (i = 0; i < 8; i ++)
    {
        t = fscanf (file, "%c", &ch);
        bb <<= 8;
        bb += ch;

        if (t == -1)
        {
            S->err = 1;
            bb = U;
            break;
        };
    };

	return bb;
};

int f_write_compact (FILE* file, U64 bb)
{
    int i;
    uchar ch;

    for (i = 0; i < 8; i ++)
    {
        ch = (bb >> (56 - i * 8)) & 0xFF;
        fprintf (file, "%c", ch);
    };

	return 0;
};

void print_this_att (U64 at)
{
	int a, b, c;
	U64 p;
	U64 t = 0;

	for (a = 0; a <= 7; a ++)
	{
		uci_send ("\n");
		for (b = 0; b <= 7; b ++)
		{
			c = compose ((7 - a), b);
			t = I << c;

			p = at & t;

			if (p) uci_send ("#");
			else   uci_send (".");
		};
	};
	uci_send ("\n");
};

int print_move (int mv)
{
	int fl, ep;
	int _from, _to, _flags, _piece;
	char id;

	_flags = (mv & 0x000000FF);
    _to    = (mv & 0x0000FF00) >> 8;
    _from  = (mv & 0x00FF0000) >> 16;
    _piece = (mv & 0xFF000000) >> 24;

	//if (_from = -1) return 0;

	uci_send (" %c", piece_array [_piece]);
	uci_send ("%c%c", 'a' + (vert (_from)), '1' + (hori (_from)));

	ep = is_ep(_flags);
	fl = is_capture(_flags) || ep;
	if (fl) uci_send ("x");

	uci_send ("%c%c", 'a' + (vert (_to)), '1' + (hori (_to)));

    id = (is_prom (_flags)) ? "nbrq"[_flags & ~f_mask] : " "[0];
    id = (ep) ? "e"[0] : id;
	uci_send ("%c", id);
	if (!fl) uci_send (" ");

	//uci_send ("\n");

	return 0;
};

void print_score (int score)
{
    uci_send ("info depth %i ", S->search_depth);

    if     (score < - MATE)
    {
        uci_send ("score mate %i ", - (INF + score) / 2);
    }
    else if (score > MATE)
    {
        uci_send ("score mate %i ",   (INF - score + 1) / 2);
    }
    else                     uci_send ("score cp %i ", score);
};

void shell_sort (int * a, int * b, int num)
{
    register int i, j, k, inc;
    int x, y;

    int arr[] = {9, 5, 3, 2, 1};

    for (k = 0; k < 5; k ++)
    {
        inc = arr[k];
        for (i = inc; i < num; ++ i)
        {
            x = a[i];
            y = b[i];

            for (j = i - inc; (x > a[j]) && (j >= 0); j -= inc)
            {
                a[j + inc] = a[j];
                b[j + inc] = b[j];
            };

            a[j + inc] = x;
            b[j + inc] = y;
        };
    };
};

int is_on_line (int a, int b)
{
    int h, v;

    h = hori (b) - hori (a);
    v = vert (b) - vert (a);

    if (h > 0)
    {
        if (v > 0) return (abs (h) == abs (v)) ? 9 : 0;
        if (v < 0) return (abs (h) == abs (v)) ? 7 : 0;
        return 8;
    };

    if (h < 0)
    {
        if (v > 0) return (abs (h) == abs (v)) ? -7 : 0;
        if (v < 0) return (abs (h) == abs (v)) ? -9 : 0;
        return -8;
    };

    if (v > 0) return 1;
    if (v < 0) return -1;

    return 0;
};

#if defined(_WIN32) || defined(_WIN64)
bool input_available ()
{
    static bool init = false, is_pipe;
    static HANDLE stdin_h;
    DWORD val;

    if (S->tuning) return 0; // don't write if we in tuning

    val = 0; // needed to make the compiler happy?

    // have a look at the "local" buffer first, *this time before init (no idea if it helps)*

    if (stdin->_cnt > 0) return true; // HACK: assumes FILE internals

    // input init (only done once)

    if (!init)
    {
        init = true;

        stdin_h = GetStdHandle(STD_INPUT_HANDLE);

        is_pipe = !GetConsoleMode(stdin_h, &val); // HACK: assumes pipe on failure

        if (!is_pipe)
        {
            SetConsoleMode (stdin_h, val & ~(ENABLE_MOUSE_INPUT|ENABLE_WINDOW_INPUT));
            FlushConsoleInputBuffer (stdin_h); // no idea if we can lose data doing this
        };
    };

    // different polling depending on input type
    // does this code work at all for pipes?

    if (is_pipe)
    {
        // HACK: assumes EOF on failure
        if (!PeekNamedPipe(stdin_h, NULL, 0, NULL, &val, NULL)) return true;

        return (val > 0); // != 0?
    }
    else
    {
        GetNumberOfConsoleInputEvents (stdin_h, &val);
        return (val > 1); // no idea why 1
    };

    return false;
};

#else

	// linux

#include <sys/select.h>

bool input_available()
{
    fd_set readfds;
    struct timeval timeout;

    FD_ZERO(&readfds);
    FD_SET(fileno(stdin), &readfds);
    /* Set to timeout immediately */
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    select(16, &readfds, 0, 0, &timeout);

    return (FD_ISSET(fileno(stdin), &readfds));
};

#endif
